"""
runpod | serverless | utils | validator.py
Provides a function to validate the input to the model.
"""

# pylint: disable=too-many-branches

import json
from typing import Any, Dict, List, Union

# Error messages
UNEXPECTED_INPUT_ERROR = "Unexpected input. {} is not a valid input option."
MISSING_REQUIRED_ERROR = "{} is a required input."
MISSING_DEFAULT_ERROR = "Schema error, missing default value for {}."
MISSING_TYPE_ERROR = "Schema error, missing type for {}."
INVALID_TYPE_ERROR = "{} should be {} type, not {}."
CONSTRAINTS_ERROR = "{} does not meet the constraints."
SCHEMA_ERROR = "Schema error, {} is not a dictionary."


def _add_error(error_list: List[str], message: str) -> None:
    error_list.append(message)


def _check_for_unexpected_inputs(raw_input, schema, error_list):
    for key in raw_input:
        if key not in schema:
            _add_error(error_list, UNEXPECTED_INPUT_ERROR.format(key))


def _validate_and_transform_schema_items(schema, error_list):
    for key, rules in schema.items():
        if not isinstance(rules, dict):
            try:
                schema[key] = json.loads(rules)
            except json.decoder.JSONDecodeError:
                _add_error(error_list, SCHEMA_ERROR.format(key))


def _validate_required_inputs_and_set_defaults(
    raw_input, schema, validated_input, error_list
):
    for key, rules in schema.items():
        if "type" not in rules:
            _add_error(error_list, MISSING_TYPE_ERROR.format(key))

        if "required" not in rules:
            _add_error(error_list, MISSING_REQUIRED_ERROR.format(key))
        elif rules["required"] and key not in raw_input:
            _add_error(error_list, MISSING_REQUIRED_ERROR.format(key))
        elif not rules["required"] and key not in raw_input:
            if "default" in rules:
                validated_input[key] = rules["default"]
            else:
                _add_error(error_list, MISSING_DEFAULT_ERROR.format(key))


def _validate_input_against_schema(schema, validated_input, error_list):
    for key, rules in schema.items():
        if key in validated_input:
            # Enforce floats to be floats.
            try:
                if rules["type"] is float and type(validated_input[key]) in [
                    int,
                    float,
                ]:
                    validated_input[key] = float(validated_input[key])
            except TypeError:
                continue

            # skip type check assuming rare case, default value sets value
            if "default" in rules and isinstance(validated_input[key], type(rules["default"])):
                continue

            # Check for the correct type.
            is_instance = isinstance(validated_input[key], rules["type"])
            if not is_instance:
                _add_error(
                    error_list,
                    f"{key} should be {rules['type']} type, not {type(validated_input[key])}.",
                )

        # Check lambda constraints.
        if "constraints" in rules and not rules["constraints"](
            validated_input.get(key)
        ):
            _add_error(error_list, CONSTRAINTS_ERROR.format(key))


def validate(
    raw_input: Dict[str, Any], schema: Dict[str, Any]
) -> Dict[str, Union[Dict[str, Any], List[str]]]:
    """
    Validates the input.
    Checks to see if the provided inputs match the expected types.
    Checks to see if the required inputs are included.
    Sets the default values for the inputs that are not provided.
    Validates the inputs using the lambda constraints.

    Returns either the list of errors or a validated_job_input.
    {"errors": ["error1", "error2"]}
    or
    {"validated_input": {"input1": "value1", "input2": "value2"}
    """
    error_list = []
    validated_input = raw_input.copy()

    # Separate the process into functions for better readability
    _check_for_unexpected_inputs(raw_input, schema, error_list)
    _validate_and_transform_schema_items(schema, error_list)
    _validate_required_inputs_and_set_defaults(
        raw_input, schema, validated_input, error_list
    )
    _validate_input_against_schema(schema, validated_input, error_list)

    validation_return = {"validated_input": validated_input}
    if error_list:
        validation_return = {"errors": error_list}

    return validation_return



"""
RunPod | API | Queries | User

Query for user information.
"""

QUERY_USER = """
query myself {
    myself {
        id
        pubKey
        networkVolumes {
            id
            name
            size
            dataCenterId
        }
    }
}
"""



""" PodWorker | modules | upload.py """

# pylint: disable=too-many-arguments

import io
import logging
import multiprocessing
import os
import shutil
import threading
import time
import uuid
from typing import Optional, Tuple
from urllib.parse import urlparse

import boto3
from boto3 import session
from boto3.s3.transfer import TransferConfig
from botocore.config import Config
from tqdm_loggable.auto import tqdm

logger = logging.getLogger("runpod upload utility")
FMT = "%(filename)-20s:%(lineno)-4d %(asctime)s %(message)s"
logging.basicConfig(level=logging.INFO, format=FMT, handlers=[logging.StreamHandler()])


def extract_region_from_url(endpoint_url):
    """
    Extracts the region from the endpoint URL.
    """
    parsed_url = urlparse(endpoint_url)
    # AWS/backblaze S3-like URL
    if ".s3." in endpoint_url:
        return endpoint_url.split(".s3.")[1].split(".")[0]

    # DigitalOcean Spaces-like URL
    if parsed_url.netloc.endswith(".digitaloceanspaces.com"):
        return endpoint_url.split(".")[1].split(".digitaloceanspaces.com")[0]

    return None


# --------------------------- S3 Bucket Connection --------------------------- #
def get_boto_client(
    bucket_creds: Optional[dict] = None,
) -> Tuple[
    boto3.client, TransferConfig
]:  # pragma: no cover # pylint: disable=line-too-long
    """
    Returns a boto3 client and transfer config for the bucket.
    """
    bucket_session = session.Session()

    boto_config = Config(
        signature_version="s3v4", retries={"max_attempts": 3, "mode": "standard"}
    )

    transfer_config = TransferConfig(
        multipart_threshold=1024 * 25,
        max_concurrency=multiprocessing.cpu_count(),
        multipart_chunksize=1024 * 25,
        use_threads=True,
    )

    if bucket_creds:
        endpoint_url = bucket_creds["endpointUrl"]
        access_key_id = bucket_creds["accessId"]
        secret_access_key = bucket_creds["accessSecret"]
    else:
        endpoint_url = os.environ.get("BUCKET_ENDPOINT_URL", None)
        access_key_id = os.environ.get("BUCKET_ACCESS_KEY_ID", None)
        secret_access_key = os.environ.get("BUCKET_SECRET_ACCESS_KEY", None)

    if endpoint_url and access_key_id and secret_access_key:
        # Extract region from the endpoint URL
        region = extract_region_from_url(endpoint_url)

        boto_client = bucket_session.client(
            "s3",
            endpoint_url=endpoint_url,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            config=boto_config,
            region_name=region,
        )
    else:
        boto_client = None

    return boto_client, transfer_config


# ---------------------------------------------------------------------------- #
#                                 Upload Image                                 #
# ---------------------------------------------------------------------------- #
def upload_image(
    job_id,
    image_location,
    result_index=0,
    results_list=None,
    bucket_name: Optional[str] = None,
):  # pylint: disable=line-too-long # pragma: no cover
    """
    Upload a single file to bucket storage.
    """
    image_name = str(uuid.uuid4())[:8]
    boto_client, _ = get_boto_client()
    file_extension = os.path.splitext(image_location)[1]
    content_type = "image/" + file_extension.lstrip(".")

    with open(image_location, "rb") as input_file:
        output = input_file.read()

    if boto_client is None:
        # Save the output to a file
        print("No bucket endpoint set, saving to disk folder 'simulated_uploaded'")
        print("If this is a live endpoint, please reference the following:")
        print(
            "https://github.com/runpod/runpod-python/blob/main/docs/serverless/utils/rp_upload.md"
        )  # pylint: disable=line-too-long

        os.makedirs("simulated_uploaded", exist_ok=True)
        sim_upload_location = f"simulated_uploaded/{image_name}{file_extension}"

        with open(sim_upload_location, "wb") as file_output:
            file_output.write(output)

        if results_list is not None:
            results_list[result_index] = sim_upload_location

        return sim_upload_location

    bucket = bucket_name if bucket_name else time.strftime("%m-%y")
    boto_client.put_object(
        Bucket=f"{bucket}",
        Key=f"{job_id}/{image_name}{file_extension}",
        Body=output,
        ContentType=content_type,
    )

    presigned_url = boto_client.generate_presigned_url(
        "get_object",
        Params={"Bucket": f"{bucket}", "Key": f"{job_id}/{image_name}{file_extension}"},
        ExpiresIn=604800,
    )

    if results_list is not None:
        results_list[result_index] = presigned_url

    return presigned_url


# ---------------------------------------------------------------------------- #
#                                Files To Upload                               #
# ---------------------------------------------------------------------------- #
def files(job_id, file_list):  # pragma: no cover
    """
    Uploads a list of files in parallel.
    Once all files are uploaded, the function returns the presigned URLs list.
    """
    upload_progress = []  # List of threads
    file_urls = [None] * len(file_list)  # Resulting list of URLs for each file

    for index, selected_file in enumerate(file_list):
        new_upload = threading.Thread(
            target=upload_image, args=(job_id, selected_file, index, file_urls)
        )

        new_upload.start()
        upload_progress.append(new_upload)

    # Wait for all uploads to finish
    for upload in upload_progress:
        upload.join()

    return file_urls


# --------------------------- Custom Bucket Upload --------------------------- #
def bucket_upload(job_id, file_list, bucket_creds):  # pragma: no cover
    """
    Uploads files to bucket storage.
    """
    temp_bucket_session = session.Session()

    temp_boto_config = Config(
        signature_version="s3v4", retries={"max_attempts": 3, "mode": "standard"}
    )

    temp_boto_client = temp_bucket_session.client(
        "s3",
        endpoint_url=bucket_creds["endpointUrl"],
        aws_access_key_id=bucket_creds["accessId"],
        aws_secret_access_key=bucket_creds["accessSecret"],
        config=temp_boto_config,
    )

    bucket_urls = []

    for selected_file in file_list:
        with open(selected_file, "rb") as file_data:
            temp_boto_client.put_object(
                Bucket=str(bucket_creds["bucketName"]),
                Key=f"{job_id}/{selected_file}",
                Body=file_data,
            )

        bucket_urls.append(
            f"{bucket_creds['endpointUrl']}/{bucket_creds['bucketName']}/{job_id}/{selected_file}"
        )

    return bucket_urls


# ------------------------- Single File Bucket Upload ------------------------ #
def upload_file_to_bucket(
    file_name: str,
    file_location: str,
    bucket_creds: Optional[dict] = None,
    bucket_name: Optional[str] = None,
    prefix: Optional[str] = None,
    extra_args: Optional[dict] = None,
) -> str:  # pragma: no cover
    """
    Uploads a single file to bucket storage and returns a presigned URL.
    """
    boto_client, transfer_config = get_boto_client(bucket_creds)

    if not bucket_name:
        bucket_name = time.strftime("%m-%y")

    key = f"{prefix}/{file_name}" if prefix else file_name

    if boto_client is None:
        print("No bucket endpoint set, saving to disk folder 'local_upload'")
        print("If this is a live endpoint, please reference the following:")
        print(
            "https://github.com/runpod/runpod-python/blob/main/docs/serverless/utils/rp_upload.md"
        )  # pylint: disable=line-too-long

        os.makedirs("local_upload", exist_ok=True)
        local_upload_location = f"local_upload/{file_name}"
        shutil.copyfile(file_location, local_upload_location)

        return local_upload_location

    file_size = os.path.getsize(file_location)
    with tqdm(
        total=file_size, unit="B", unit_scale=True, desc=file_name
    ) as progress_bar:
        upload_file_args = {
            "Filename": file_location,
            "Bucket": bucket_name,
            "Key": key,
            "Config": transfer_config,
            "Callback": progress_bar.update,
        }

        if extra_args:
            upload_file_args["ExtraArgs"] = extra_args

        boto_client.upload_file(**upload_file_args)

    presigned_url = boto_client.generate_presigned_url(
        "get_object", Params={"Bucket": bucket_name, "Key": key}, ExpiresIn=604800
    )

    return presigned_url


# --------------------------- Upload Memory Object --------------------------- #
def upload_in_memory_object(
    file_name: str,
    file_data: bytes,
    bucket_creds: Optional[dict] = None,
    bucket_name: Optional[str] = None,
    prefix: Optional[str] = None,
) -> str:  # pragma: no cover
    """
    Uploads an in-memory object (bytes) to bucket storage and returns a presigned URL.
    """
    boto_client, transfer_config = get_boto_client(bucket_creds)

    if not bucket_name:
        bucket_name = time.strftime("%m-%y")

    key = f"{prefix}/{file_name}" if prefix else file_name

    file_size = len(file_data)
    with tqdm(
        total=file_size, unit="B", unit_scale=True, desc=file_name
    ) as progress_bar:
        boto_client.upload_fileobj(
            io.BytesIO(file_data),
            bucket_name,
            key,
            Config=transfer_config,
            Callback=progress_bar.update,
        )

    presigned_url = boto_client.generate_presigned_url(
        "get_object", Params={"Bucket": bucket_name, "Key": key}, ExpiresIn=604800
    )

    return presigned_url



""" Tests for the user_agent module. """

import os
import unittest
from unittest.mock import patch

from runpod import __version__ as runpod_version
from runpod.user_agent import construct_user_agent


class TestConstructUserAgent(unittest.TestCase):
    """Test the construct_user_agent function."""

    @patch("runpod.user_agent.platform.system", return_value="Windows")
    @patch("runpod.user_agent.platform.release", return_value="10")
    @patch("runpod.user_agent.platform.machine", return_value="AMD64")
    @patch("runpod.user_agent.platform.python_version", return_value="3.8.10")
    def test_user_agent_without_integration(
        self, mock_python_version, mock_machine, mock_release, mock_system
    ):
        """Test the User-Agent string without specifying an integration method."""
        if "RUNPOD_UA_INTEGRATION" in os.environ:
            del os.environ["RUNPOD_UA_INTEGRATION"]

        expected_ua = f"RunPod-Python-SDK/{runpod_version} (Windows 10; AMD64) Language/Python 3.8.10"  # pylint: disable=line-too-long
        self.assertEqual(construct_user_agent(), expected_ua)

        assert mock_python_version.called
        assert mock_machine.called
        assert mock_release.called
        assert mock_system.called

    @patch("runpod.user_agent.platform.system", return_value="Linux")
    @patch("runpod.user_agent.platform.release", return_value="5.4")
    @patch("runpod.user_agent.platform.machine", return_value="x86_64")
    @patch("runpod.user_agent.platform.python_version", return_value="3.9.5")
    @patch.dict(os.environ, {"RUNPOD_UA_INTEGRATION": "SkyPilot"})
    def test_user_agent_with_integration(
        self, mock_python_version, mock_machine, mock_release, mock_system
    ):
        """Test the User-Agent string with an integration method specified."""
        expected_ua = f"RunPod-Python-SDK/{runpod_version} (Linux 5.4; x86_64) Language/Python 3.9.5 Integration/SkyPilot"  # pylint: disable=line-too-long

        os.environ["RUNPOD_UA_INTEGRATION"] = "SkyPilot"
        self.assertEqual(construct_user_agent(), expected_ua)
        os.environ.pop("RUNPOD_UA_INTEGRATION")

        assert mock_python_version.called
        assert mock_machine.called
        assert mock_release.called
        assert mock_system.called


if __name__ == "__main__":
    unittest.main()



"""Helper functions for the runpod cli group exec command"""

import os

import click

from runpod import get_pod

POD_ID_FILE = os.path.join(os.path.expanduser("~"), ".runpod", "pod_id")


def get_session_pod():
    """Returns the pod_id for the current session.

    Session pod is used to execute commands and run scripts remotely.

    - If the pod_id is already set, return it.
    - If the pod_id is not set, prompt the user for it.
    - Save the pod_id to a file so that the user doesn't have to enter it again.
    """
    pod_id = None

    if os.path.exists(POD_ID_FILE):
        with open(POD_ID_FILE, "r", encoding="UTF-8") as pod_file:
            pod_id = pod_file.read().strip()

    # Confirm that the pod_id is valid
    if get_pod(pod_id) is not None:
        return pod_id

    # If file doesn't exist or is empty, prompt user for the pod_id
    pod_id = click.prompt("Please provide the pod ID")
    os.makedirs(os.path.dirname(POD_ID_FILE), exist_ok=True)
    with open(POD_ID_FILE, "w", encoding="UTF-8") as pod_file:
        pod_file.write(pod_id)

    return pod_id



"""
RunPod | CLI | Utils | SSH Command
"""

import unittest
from unittest.mock import MagicMock, patch

import paramiko

from runpod.cli.utils.ssh_cmd import SSHConnection


class TestSSHConnection(unittest.TestCase):
    """Test the SSHConnection class."""

    def setUp(self):

        self.patch_get_pod_ssh_ip_port = patch(
            "runpod.cli.utils.ssh_cmd.get_pod_ssh_ip_port",
            return_value=("127.0.0.1", 22),
        ).start()

        self.patch_find_ssh_key_file = patch(
            "runpod.cli.utils.ssh_cmd.find_ssh_key_file", return_value="key_file"
        ).start()

        self.mock_ssh_client = MagicMock()
        patch_paramiko = patch(
            "runpod.cli.utils.ssh_cmd.paramiko.SSHClient",
            return_value=self.mock_ssh_client,
        ).start()

        self.addCleanup(self.patch_get_pod_ssh_ip_port.stop)
        self.addCleanup(self.patch_find_ssh_key_file.stop)
        self.addCleanup(patch_paramiko.stop)

        self.ssh_connection = SSHConnection("pod_id_mock")

    def test_enter(self):
        """Test entering the context manager."""
        self.assertEqual(
            self.ssh_connection, self.ssh_connection.__enter__()
        )  # pylint: disable=unnecessary-dunder-call

    def test_enter_exception(self):
        """Test entering the context manager with an exception."""
        self.mock_ssh_client.connect.side_effect = paramiko.SSHException
        with self.assertRaises(SystemExit):
            SSHConnection("pod_id_mock")

    def test_exit(self):
        """Test exiting the context manager."""
        self.ssh_connection.__exit__(None, None, None)

        self.mock_ssh_client.close.assert_called_once()

    def test_run_commands(self):
        """Test that run_commands() calls exec_command() on the SSH object."""
        commands = ["command1", "command2"]

        mock_exec_command = self.mock_ssh_client.exec_command
        mock_exec_command.return_value = (
            None,
            ["stdout1", "stdout2"],
            ["stderr1", "stderr2"],
        )
        self.ssh_connection.run_commands(commands)

        assert mock_exec_command.call_count == 2

    def test_put_file(self):
        """Test that put_file() calls put() on the SFTP object."""
        local_path = "/local/file.txt"
        remote_path = "/remote/file.txt"

        mock_sftp = self.mock_ssh_client.open_sftp.return_value.__enter__.return_value
        self.ssh_connection.put_file(local_path, remote_path)

        mock_sftp.put.assert_called_once_with(local_path, remote_path)

    def test_get_file(self):
        """Test that get_file() calls get() on the SFTP object."""
        local_path = "/local/file.txt"
        remote_path = "/remote/file.txt"

        mock_sftp = self.mock_ssh_client.open_sftp.return_value.__enter__.return_value
        self.ssh_connection.get_file(remote_path, local_path)

        mock_sftp.get.assert_called_once_with(remote_path, local_path)

    @patch("subprocess.run")
    def test_launch_terminal(self, mock_subprocess):
        """Test that launch_terminal() calls subprocess.run()."""
        self.ssh_connection.launch_terminal()
        mock_subprocess.assert_called_once()

    @patch("subprocess.run")
    def test_rsync(self, mock_subprocess):
        """Test that rsync() calls subprocess.run()."""
        self.ssh_connection.rsync("local_path", "remote_path", quiet=True)
        mock_subprocess.assert_called_once()

    # Test that the signal handler closes the connection.
    @patch("runpod.cli.utils.ssh_cmd.SSHConnection.close")
    def test_signal_handler(self, mock_close):
        """Test that the signal handler closes the connection."""
        with patch("sys.exit") as mock_exit:
            self.ssh_connection._signal_handler(
                None, None
            )  # pylint: disable=protected-access
        mock_close.assert_called_once()
        assert mock_exit.called



""" RunPod | API Wrapper | Mutations | Endpoints """

# pylint: disable=too-many-arguments


def generate_endpoint_mutation(
    name: str,
    template_id: str,
    gpu_ids: str = "AMPERE_16",
    network_volume_id: str = None,
    locations: str = None,
    idle_timeout: int = 5,
    scaler_type: str = "QUEUE_DELAY",
    scaler_value: int = 4,
    workers_min: int = 0,
    workers_max: int = 3,
    flashboot=False,
    allowed_cuda_versions: str = "12.1,12.2,12.3,12.4,12.5",
    gpu_count: int = None,
):
    """Generate a string for a GraphQL mutation to create a new endpoint."""
    input_fields = []

    # ------------------------------ Required Fields ----------------------------- #
    if flashboot:
        name = name + "-fb"

    input_fields.append(f'name: "{name}"')
    input_fields.append(f'templateId: "{template_id}"')
    input_fields.append(f'gpuIds: "{gpu_ids}"')

    # ------------------------------ Optional Fields ----------------------------- #
    if network_volume_id is not None:
        input_fields.append(f'networkVolumeId: "{network_volume_id}"')
    else:
        input_fields.append('networkVolumeId: ""')

    if locations is not None:
        input_fields.append(f'locations: "{locations}"')
    else:
        input_fields.append('locations: ""')

    input_fields.append(f"idleTimeout: {idle_timeout}")
    input_fields.append(f'scalerType: "{scaler_type}"')
    input_fields.append(f"scalerValue: {scaler_value}")
    input_fields.append(f"workersMin: {workers_min}")
    input_fields.append(f"workersMax: {workers_max}")

    if allowed_cuda_versions is not None:
        input_fields.append(f'allowedCudaVersions: "{allowed_cuda_versions}"')
    
    if gpu_count is not None:
        input_fields.append(f"gpuCount: {gpu_count}")

    # Format the input fields into a string
    input_fields_string = ", ".join(input_fields)

    return f"""
    mutation {{
        saveEndpoint(
            input: {{
                {input_fields_string}
            }}
        ) {{
            id
            name
            templateId
            gpuIds
            networkVolumeId
            locations
            idleTimeout
            scalerType
            scalerValue
            workersMin
            workersMax
            allowedCudaVersions
            gpuCount
        }}
    }}
    """



def update_endpoint_template_mutation(endpoint_id: str, template_id: str):
    """Generate a string for a GraphQL mutation to update an existing endpoint's template."""
    input_fields = []

    # ------------------------------ Required Fields ----------------------------- #
    input_fields.append(f'templateId: "{template_id}"')
    input_fields.append(f'endpointId: "{endpoint_id}"')

    # Format the input fields into a string
    input_fields_string = ", ".join(input_fields)
    result = f"""
    mutation {{
        updateEndpointTemplate(input: {{{input_fields_string}}}) {{
            id
            templateId
        }}
    }}
    """
    return result



"""
RunPod | CLI | Utils | SSH Command

Connect and run commands over SSH.
"""

import signal
import subprocess
import sys
import threading

import colorama
import paramiko

from runpod.cli import STOP_EVENT

from .rp_info import get_pod_ssh_ip_port
from .rp_runpodignore import get_ignore_list
from .rp_userspace import find_ssh_key_file

colorama.init(autoreset=True)  # Initialize colorama


class SSHConnection:
    """Connect and run commands over SSH."""

    def __init__(self, pod_id):
        self.pod_id = pod_id

        try:
            self.pod_ip, self.pod_port = get_pod_ssh_ip_port(pod_id)
            self.key_file = find_ssh_key_file(self.pod_ip, self.pod_port)

            self.ssh = paramiko.SSHClient()
            self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            self.ssh.connect(
                self.pod_ip,
                port=self.pod_port,
                username="root",
                key_filename=self.key_file,
            )
        except paramiko.SSHException as err:
            print(colorama.Fore.RED + f"[{pod_id}]", err)
            sys.exit(1)

        signal.signal(signal.SIGINT, self._signal_handler)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        self.close()

    def _get_ssh_options(self):
        """Get the SSH options for connecting to the pod."""
        return [
            "-o",
            "StrictHostKeyChecking=no",
            "-o",
            "LogLevel=ERROR",
            "-p",
            str(self.pod_port),
            "-i",
            self.key_file,
        ]

    def _signal_handler(self, signum, frame):
        """Handle signals."""
        del signum, frame
        self.close()
        print(colorama.Fore.BLUE + f"[{self.pod_id}]", "SSH Connection Closed")
        STOP_EVENT.set()
        sys.exit(0)

    def run_commands(self, commands):
        """Runs a list of bash commands over SSH."""

        def handle_stream(stream, color, prefix):
            for line in stream:
                if line:
                    print(color + f"[{prefix}]", line.strip())

        for command in commands:
            full_command = " && ".join(
                [
                    "source /root/.bashrc",
                    "source /etc/rp_environment",
                    "while IFS= read -r -d '' line; do export \"$line\"; done < /proc/1/environ",
                    command,
                ]
            )
            _, stdout, stderr = self.ssh.exec_command(full_command)

            stdout_thread = threading.Thread(
                target=handle_stream,
                args=(stdout, colorama.Fore.GREEN, self.pod_id),
                daemon=True,
            )
            stderr_thread = threading.Thread(
                target=handle_stream,
                args=(stderr, colorama.Fore.RED, self.pod_id),
                daemon=True,
            )

            stdout_thread.start()
            stderr_thread.start()

            stdout_thread.join()
            stderr_thread.join()

    def put_file(self, local_path, remote_path):
        """Copy local file to remote machine over SSH."""
        with self.ssh.open_sftp() as sftp:
            sftp.put(local_path, remote_path)

    def get_file(self, remote_path, local_path):
        """Fetch a remote file to local machine over SSH."""
        with self.ssh.open_sftp() as sftp:
            sftp.get(remote_path, local_path)

    def launch_terminal(self):
        """Launch an interactive terminal over SSH."""
        cmd = ["ssh"] + self._get_ssh_options() + [f"root@{self.pod_ip}"]

        subprocess.run(cmd, check=True)

    def rsync(self, local_path, remote_path, quiet=False):
        """Sync a local directory to a remote directory over SSH.

        A .runpodignore file can be used to ignore files and directories.
        This file should be placed in the root of the local directory to sync.

        Args:
            local_path (str): The local directory to sync.
            remote_path (str): The remote directory to sync.
        """
        rsync_cmd = ["rsync", "-avz", "--no-owner", "--no-group"]

        for pattern in get_ignore_list():
            rsync_cmd.extend(["--exclude", pattern])

        if quiet:
            rsync_cmd.append("--quiet")

        rsync_cmd.extend(
            [
                "-e",
                f"ssh {' '.join(self._get_ssh_options())}",
                local_path,
                f"root@{self.pod_ip}:{remote_path}",
            ]
        )

        return subprocess.run(rsync_cmd, check=True)

    def close(self):
        """Close the SSH connection."""
        self.ssh.close()



"""Unit tests for the asyncio_runner module."""

# pylint: disable=too-few-public-methods

import asyncio
import tracemalloc
import unittest
from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock, MagicMock, patch

from runpod.endpoint.asyncio.asyncio_runner import Endpoint, Job

tracemalloc.start()


class TestJob(IsolatedAsyncioTestCase):
    """Tests the Job class."""

    async def test_status(self):
        """
        Tests Job.status
        """
        with patch(
            "aiohttp.ClientSession", new_callable=AsyncMock
        ) as mock_session_class:
            mock_session = mock_session_class.return_value
            mock_get = mock_session.get
            mock_resp = AsyncMock()

            mock_resp.json.return_value = {"status": "COMPLETED"}
            mock_get.return_value = mock_resp

            job = Job("endpoint_id", "job_id", mock_session)
            status = await job.status()
            assert status == "COMPLETED"
            assert await job.status() == "COMPLETED"

    async def test_output(self):
        """
        Tests Job.output
        """
        with patch(
            "runpod.endpoint.asyncio.asyncio_runner.asyncio.sleep"
        ) as mock_sleep, patch(
            "aiohttp.ClientSession", new_callable=AsyncMock
        ) as mock_session_class:
            mock_session = mock_session_class.return_value
            mock_get = mock_session.get
            mock_resp = AsyncMock()

            async def json_side_effect():
                if mock_sleep.call_count == 0:
                    return {"status": "IN_PROGRESS"}
                return {"output": "OUTPUT", "status": "COMPLETED"}

            mock_resp.json.side_effect = json_side_effect
            mock_get.return_value = mock_resp

            job = Job("endpoint_id", "job_id", mock_session)
            output_task = asyncio.create_task(job.output(timeout=3))

            output = await output_task
            assert output == "OUTPUT"
            assert await job.output() == "OUTPUT"

    async def test_output_timeout(self):
        """
        Tests Job.output with a timeout
        """
        with patch(
            "aiohttp.ClientSession", new_callable=AsyncMock
        ) as mock_session_class:
            mock_session = mock_session_class.return_value
            mock_get = mock_session.get
            mock_resp = AsyncMock()

            mock_resp.json.return_value = {"status": "IN_PROGRESS"}
            mock_get.return_value = mock_resp

            job = Job("endpoint_id", "job_id", mock_session)
            output_task = asyncio.create_task(job.output(timeout=1))

            with self.assertRaises(TimeoutError):
                await output_task

    async def test_stream(self):
        """
        Tests Job.stream
        """
        with patch(
            "aiohttp.ClientSession", new_callable=AsyncMock
        ) as mock_session_class:
            mock_session = mock_session_class.return_value
            mock_get = mock_session.get
            mock_resp = AsyncMock()

            responses = [
                {"stream": [{"output": "OUTPUT1"}], "status": "IN_PROGRESS"},
                {"stream": [{"output": "OUTPUT2"}], "status": "IN_PROGRESS"},
            ]

            async def json_side_effect():
                return (
                    responses.pop(0)
                    if responses
                    else {"stream": [], "status": "COMPLETED"}
                )

            mock_resp.json.side_effect = json_side_effect
            mock_get.return_value = mock_resp

            job = Job("endpoint_id", "job_id", mock_session)

            outputs = []
            async for stream_output in job.stream():
                outputs.append(stream_output)

            assert outputs == ["OUTPUT1", "OUTPUT2"]

    async def test_cancel(self):
        """
        Tests Job.cancel
        """
        with patch("aiohttp.ClientSession") as mock_session:
            mock_resp = MagicMock()
            mock_resp.json = MagicMock(return_value=asyncio.Future())
            mock_resp.json.return_value.set_result({"result": "CANCELLED"})
            mock_session.post.return_value.__aenter__.return_value = mock_resp

            job = Job("endpoint_id", "job_id", mock_session)
            cancel_result = await job.cancel()
            assert cancel_result == {"result": "CANCELLED"}

    async def test_output_in_progress_then_completed(self):
        """Tests Job.output when status is initially IN_PROGRESS and then changes to COMPLETED"""
        with patch(
            "aiohttp.ClientSession", new_callable=AsyncMock
        ) as mock_session_class:
            mock_session = mock_session_class.return_value
            mock_get = mock_session.get
            mock_resp = AsyncMock()

            responses = [
                {"status": "IN_PROGRESS"},
                {"status": "COMPLETED", "output": "OUTPUT"},
            ]

            async def json_side_effect():
                return (
                    responses.pop(0)
                    if responses
                    else {"status": "COMPLETED", "output": "OUTPUT"}
                )  # pylint: disable=line-too-long

            mock_resp.json.side_effect = json_side_effect
            mock_get.return_value = mock_resp

            job = Job("endpoint_id", "job_id", mock_session)
            output = await job.output(timeout=3)
            assert output == "OUTPUT"


class TestEndpoint(IsolatedAsyncioTestCase):
    """Unit tests for the Endpoint class."""

    async def test_run(self):
        """
        Tests Endpoint.run
        """
        with patch("aiohttp.ClientSession") as mock_session:
            mock_resp = MagicMock()
            mock_resp.json = MagicMock(return_value=asyncio.Future())
            mock_resp.json.return_value.set_result({"id": "job_id"})
            mock_session.post.return_value.__aenter__.return_value = mock_resp

            endpoint = Endpoint("endpoint_id", mock_session)
            job = await endpoint.run({"input": "INPUT"})
            assert job.job_id == "job_id"

    async def test_health(self):
        """
        Tests Endpoint.health
        """
        with patch("aiohttp.ClientSession") as mock_session:
            mock_resp = MagicMock()
            mock_resp.json = MagicMock(return_value=asyncio.Future())
            mock_resp.json.return_value.set_result({"status": "HEALTHY"})
            mock_session.get.return_value.__aenter__.return_value = mock_resp

            endpoint = Endpoint("endpoint_id", mock_session)
            health = await endpoint.health()
            assert health == {"status": "HEALTHY"}

    async def test_purge_queue(self):
        """
        Tests Endpoint.purge_queue
        """
        with patch("aiohttp.ClientSession") as mock_session:
            mock_resp = MagicMock()
            mock_resp.json = MagicMock(return_value=asyncio.Future())
            mock_resp.json.return_value.set_result({"result": "PURGED"})
            mock_session.post.return_value.__aenter__.return_value = mock_resp

            endpoint = Endpoint("endpoint_id", mock_session)
            purge_result = await endpoint.purge_queue()
            assert purge_result == {"result": "PURGED"}


class TestEndpointInitialization(unittest.TestCase):
    """Tests for the Endpoint class initialization."""

    def test_endpoint_initialization(self):
        """Tests initialization of Endpoint class."""
        with patch("aiohttp.ClientSession"):
            endpoint = Endpoint("endpoint_id", MagicMock())
            self.assertEqual(
                endpoint.endpoint_url, "https://api.runpod.ai/v2/endpoint_id/run"
            )
            self.assertIn("Content-Type", endpoint.headers)
            self.assertIn("Authorization", endpoint.headers)


if __name__ == "__main__":
    unittest.main()

    snapshot = tracemalloc.take_snapshot()
    top_stats = snapshot.statistics("lineno")

    print("[ Top 10 ]")
    for stat in top_stats[:10]:
        print(stat)



"""
RunPod | CLI | Utils | Userspace
"""

import os

import paramiko

from runpod import SSH_KEY_PATH


def find_ssh_key_file(pod_ip, pod_port):
    """Find the SSH key file that can be used to connect to the pod.

    - Try all the keys in the SSH_KEY_PATH directory
    - If none of the keys work, return None
    - If multiple keys work, return the first one that works
    - Returns the path to the key file
    """
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    for file in os.listdir(SSH_KEY_PATH):
        file_path = os.path.join(SSH_KEY_PATH, file)

        if not os.path.isfile(file_path) or file.endswith(".pub"):
            continue

        try:
            ssh.connect(pod_ip, port=pod_port, username="root", key_filename=file_path)
            ssh.close()
            print(f"Connected to pod {pod_ip}:{pod_port} using key {file}")
            return file_path
        except Exception as err:  # pylint: disable=broad-except
            print(f"An error occurred with key {file}: {err}")

    print("Failed to connect using all available keys.")
    return None



""" Unit tests for the runpod.cli.groups.exec.helpers module. """

import unittest
from unittest.mock import mock_open, patch

from runpod.cli.groups.exec.helpers import POD_ID_FILE, get_session_pod


class TestGetSessionPod(unittest.TestCase):
    """Unit tests for get_session_pod"""

    def setUp(self):
        self.mocked_pod_id = "sample_pod_id"

    @patch("os.path.exists")
    @patch("runpod.cli.groups.exec.helpers.get_pod")
    def test_existing_pod_id_file_and_valid_pod_id(self, mock_get_pod, mock_exists):
        """Test get_session_pod when the pod_id file exists and the pod_id is valid"""
        mock_exists.return_value = True
        mock_get_pod.return_value = True
        with patch("builtins.open", mock_open(read_data=self.mocked_pod_id)):
            result = get_session_pod()
        self.assertEqual(result, self.mocked_pod_id)

    @patch("os.path.exists")
    @patch("runpod.cli.groups.exec.helpers.get_pod")
    def test_existing_pod_id_file_and_invalid_pod_id(self, mock_get_pod, mock_exists):
        """Test get_session_pod when the pod_id file exists and the pod_id is invalid"""
        mock_exists.return_value = True
        mock_get_pod.return_value = None
        with patch("builtins.open", mock_open(read_data="invalid_pod_id")):
            with patch("click.prompt", return_value=self.mocked_pod_id):
                result = get_session_pod()
        self.assertEqual(result, self.mocked_pod_id)

    @patch("os.path.exists")
    @patch("runpod.cli.groups.exec.helpers.get_pod")
    def test_no_pod_id_file(self, mock_get_pod, mock_exists):
        """Test get_session_pod when the pod_id file doesn't exist"""
        mock_exists.return_value = False
        mock_get_pod.return_value = None
        with patch("click.prompt", return_value=self.mocked_pod_id):
            result = get_session_pod()
        self.assertEqual(result, self.mocked_pod_id)

    @patch("os.path.exists")
    @patch("runpod.cli.groups.exec.helpers.get_pod")
    def test_pod_id_file_written_to_when_not_existing(self, mock_get_pod, mock_exists):
        """Test get_session_pod when the pod_id file doesn't exist"""
        mock_exists.return_value = False
        mock_get_pod.return_value = None
        mocked = mock_open()
        with patch("builtins.open", mocked):
            with patch("click.prompt", return_value=self.mocked_pod_id):
                get_session_pod()
        mocked.assert_called_once_with(POD_ID_FILE, "w", encoding="UTF-8")
        handle = mocked()
        handle.write.assert_called_once_with(self.mocked_pod_id)

    @patch("os.path.exists")
    @patch("runpod.cli.groups.exec.helpers.get_pod")
    def test_pod_id_file_written_to_when_invalid_pod_id_in_file(
        self, mock_get_pod, mock_exists
    ):
        """Test get_session_pod when the pod_id file exists and the pod_id is invalid"""
        mock_exists.return_value = True
        mock_get_pod.return_value = None
        mocked = mock_open(read_data="invalid_pod_id")
        with patch("builtins.open", mocked):
            with patch("click.prompt", return_value=self.mocked_pod_id):
                get_session_pod()
        mocked.assert_called_with(POD_ID_FILE, "w", encoding="UTF-8")
        handle = mocked()
        handle.write.assert_called_with(self.mocked_pod_id)



""" Tests for runpod | serverless| worker """

# pylint: disable=protected-access

import argparse
import os
import sys
from unittest import mock
from unittest.mock import patch, mock_open, Mock, MagicMock

from unittest import IsolatedAsyncioTestCase
import nest_asyncio

import runpod
from runpod.serverless.modules.rp_logger import RunPodLogger
from runpod.serverless.modules.rp_scale import _handle_uncaught_exception
from runpod.serverless import _signal_handler

nest_asyncio.apply()


class TestWorker(IsolatedAsyncioTestCase):
    """Tests for RunPod serverless worker."""

    async def asyncSetUp(self):
        self.mock_handler = mock.Mock(return_value="test")
        self.mock_config = {
            "handler": self.mock_handler,
            "rp_args": {"test_input": None},
        }

    def test_is_local(self):
        """
        Test _is_local
        """
        with patch("runpod.serverless.worker.os") as mock_os:
            mock_os.environ.get.return_value = None
            assert runpod.serverless.worker._is_local({"rp_args": {}}) is True
            assert (
                runpod.serverless.worker._is_local(
                    {"rp_args": {"test_input": "something"}}
                )
                is True
            )

            mock_os.environ.get.return_value = "something"
            assert runpod.serverless.worker._is_local(self.mock_config) is False

    def test_start(self):
        """
        Test basic start call.
        """
        with patch(
            "builtins.open", mock_open(read_data='{"input":{"number":1}}')
        ) as mock_file, self.assertRaises(SystemExit):

            runpod.serverless.start({"handler": self.mock_handler})

            assert mock_file.called

    def test_is_local_testing(self):
        """
        Test _is_local_testing
        """
        with patch("runpod.serverless.worker.os") as mock_os:
            mock_os.environ.get.return_value = None
            assert runpod.serverless.worker._is_local(self.mock_config) is True
            mock_os.environ.get.return_value = "something"
            assert runpod.serverless.worker._is_local(self.mock_config) is False

    def test_local_api(self):
        """
        Test local FastAPI setup.
        """
        known_args = argparse.Namespace()
        known_args.rp_log_level = None
        known_args.rp_debugger = None
        known_args.rp_serve_api = True
        known_args.rp_api_port = 8000
        known_args.rp_api_concurrency = 1
        known_args.rp_api_host = "localhost"
        known_args.test_input = '{"test": "test"}'

        with patch(
            "argparse.ArgumentParser.parse_known_args"
        ) as mock_parse_known_args, patch(
            "runpod.serverless.rp_fastapi"
        ) as mock_fastapi:

            mock_parse_known_args.return_value = known_args, []
            runpod.serverless.start({"handler": self.mock_handler})

            assert mock_fastapi.WorkerAPI.called

    @patch("runpod.serverless.log")
    @patch("runpod.serverless.sys.exit")
    def test_signal_handler(self, mock_exit, mock_logger):
        """
        Test signal handler.
        """

        _signal_handler(None, None)

        assert mock_exit.called
        assert mock_logger.info.called


class TestWorkerTestInput(IsolatedAsyncioTestCase):
    """Tests for runpod | serverless| worker"""

    async def asyncSetUp(self):
        self.mock_handler = Mock()
        self.mock_handler.return_value = {}

        self.mock_handler.return_value = "test"

    def test_worker_bad_local(self):
        """
        Test sys args.
        """
        known_args = argparse.Namespace()
        known_args.rp_log_level = "WARN"
        known_args.rp_debugger = True
        known_args.rp_serve_api = None
        known_args.rp_api_port = 8000
        known_args.rp_api_concurrency = 1
        known_args.rp_api_host = "localhost"
        known_args.test_input = '{"test": "test"}'
        known_args.test_output = '{"test": "test"}'

        with patch(
            "argparse.ArgumentParser.parse_known_args"
        ) as mock_parse_known_args, self.assertRaises(SystemExit):

            mock_parse_known_args.return_value = known_args, []
            runpod.serverless.start({"handler": self.mock_handler})

            # Confirm that the log level is set to WARN
            log = RunPodLogger()
            assert log.level == "WARN"


def generator_handler(job):
    """
    Test generator_handler
    """
    print(job)
    yield "test1"
    yield "test2"


def generator_handler_exception(job):
    """
    Test generator_handler
    """
    print(job)
    yield "test1"
    print("Raise exception")
    raise Exception()  # pylint: disable=broad-exception-raised


def test_generator_handler_exception():
    """Test generator_handler_exception"""
    job = {"id": "test_job"}
    gen = generator_handler_exception(job)

    # Process the first yielded value
    output = next(gen)
    assert output == "test1", "First output should be 'test1'"

    # Attempt to get the next value, expecting an exception
    try:
        next(gen)
        assert False, "Expected an exception to be raised"
    except Exception:  # pylint: disable=broad-except
        assert True, "Exception was caught as expected"


class TestRunWorker(IsolatedAsyncioTestCase):
    """Tests for runpod | serverless| worker"""

    async def asyncSetUp(self):
        os.environ["RUNPOD_WEBHOOK_GET_JOB"] = "https://test.com"

        # Set up the config
        self.config = {
            "handler": MagicMock(),
            "refresh_worker": True,
            "rp_args": {"rp_debugger": True, "rp_log_level": "DEBUG"},
        }

    async def asyncTearDown(self):
        sys.excepthook = sys.__excepthook__

    @patch("runpod.serverless.modules.rp_scale.AsyncClientSession")
    @patch("runpod.serverless.modules.rp_scale.get_job")
    @patch("runpod.serverless.modules.rp_job.run_job")
    @patch("runpod.serverless.modules.rp_job.stream_result")
    @patch("runpod.serverless.modules.rp_job.send_result")
    async def test_run_worker(
        self,
        mock_send_result,
        mock_stream_result,
        mock_run_job,
        mock_get_job,
        mock_session,
    ):
        """Test run_worker with synchronous handler."""
        # Mock return values for get_job
        mock_get_job.side_effect = [
            [{"id": "123", "input": {"number": 1}}],
            []  # Stop the loop after the second call
        ]
        mock_run_job.return_value = {"output": {"result": "odd"}}

        # Call the function
        runpod.serverless.start(self.config)

        # Make assertions about the behaviors
        self.assertEqual(mock_get_job.call_count, 1)
        mock_run_job.assert_called_once()
        mock_send_result.assert_called_once()

        assert not mock_stream_result.called
        assert mock_session.called

    @patch("runpod.serverless.modules.rp_scale.get_job")
    @patch("runpod.serverless.modules.rp_job.run_job")
    @patch("runpod.serverless.modules.rp_job.stream_result")
    @patch("runpod.serverless.modules.rp_job.send_result")
    async def test_run_worker_generator_handler(
        self, mock_send_result, mock_stream_result, mock_run_job, mock_get_job
    ):
        """
        Test run_worker with generator handler.

        Args:
            mock_stream_result (_type_): _description_
            mock_run_job_generator (_type_): _description_
            mock_run_job (_type_): _description_
            mock_get_job (_type_): _description_
        """
        # Define the mock behaviors
        mock_get_job.return_value = [{"id": "generator-123", "input": {"number": 1}}]

        # Test generator handler
        generator_config = {"handler": generator_handler, "refresh_worker": True}
        runpod.serverless.start(generator_config)

        assert mock_stream_result.called
        assert not mock_run_job.called

        # Since return_aggregate_stream is NOT activated, we should not submit any outputs.
        _, args, _ = mock_send_result.mock_calls[0]
        assert args[1] == {"output": [], "stopPod": True}

    @patch("runpod.serverless.modules.rp_scale.get_job")
    @patch("runpod.serverless.modules.rp_job.run_job")
    @patch("runpod.serverless.modules.rp_job.stream_result")
    @patch("runpod.serverless.modules.rp_job.send_result")
    async def test_run_worker_generator_handler_exception(
        self, mock_send_result, mock_stream_result, mock_run_job, mock_get_job
    ):
        """
        Test run_worker with generator handler.

        This test verifies that:
        - `stream_result` is called before an exception occurs.
        - `run_job` is never called since `handler` is a generator function.
        - An error is correctly reported back via `send_result`.
        """
        RunPodLogger().set_level("DEBUG")

        # Setup: Mock `get_job` to return a predefined job.
        mock_get_job.return_value = [
            {"id": "generator-123-exception", "input": {"number": 1}}
        ]

        runpod.serverless.start(
            {"handler": generator_handler_exception, "refresh_worker": True}
        )

        # Ensure `stream_result` was called at least once
        assert mock_stream_result.call_count >= 1

        # Ensure `run_job` was not called since the handler is a generator function
        assert not mock_run_job.called

        # Check that `send_result` was called
        assert mock_send_result.call_count == 1  # Adjust expectation if multiple calls are valid

        # Inspect the arguments for each call to `send_result`
        for call in mock_send_result.call_args_list:
            args, kwargs = call  # Unpack the tuple into args and kwargs
            # Check if the expected key is present in the args or kwargs
            if args and len(args) > 1:
                assert "error" in args[1] or "result" in args[1], "Expected error or result in args."
            else:
                # If args[1] doesn't have the expected keys, check in kwargs
                assert "error" in kwargs or "result" in kwargs, "Expected error or result in kwargs."

    @patch("runpod.serverless.modules.rp_scale.get_job")
    @patch("runpod.serverless.modules.rp_job.run_job")
    @patch("runpod.serverless.modules.rp_job.stream_result")
    @patch("runpod.serverless.modules.rp_job.send_result")
    async def test_run_worker_generator_aggregate_handler(
        self, mock_send_result, mock_stream_result, mock_run_job, mock_get_job
    ):
        """
        Test run_worker with generator handler.

        Args:
            mock_send_result (_type_): _description_
            mock_stream_result (_type_): _description_
            mock_run_job (_type_): _description_
            mock_get_job (_type_): _description_
            mock_session (_type_): _description_
        """
        # Define the mock behaviors
        mock_get_job.return_value = [
            {"id": "generator-123-aggregate", "input": {"number": 1}}
        ]

        # Test generator handler
        generator_config = {
            "handler": generator_handler,
            "return_aggregate_stream": True,
            "refresh_worker": True,
        }

        runpod.serverless.start(generator_config)

        assert mock_send_result.called
        assert mock_stream_result.called
        assert not mock_run_job.called

        # Since return_aggregate_stream is activated, we should submit a list of the outputs.
        _, args, _ = mock_send_result.mock_calls[0]
        assert args[1] == {"output": ["test1", "test2"], "stopPod": True}

    @patch("runpod.serverless.modules.rp_scale.AsyncClientSession")
    @patch("runpod.serverless.modules.rp_scale.get_job")
    @patch("runpod.serverless.modules.rp_job.run_job")
    @patch("runpod.serverless.modules.rp_job.stream_result")
    @patch("runpod.serverless.modules.rp_job.send_result")
    async def test_run_worker_concurrency(
        self,
        mock_send_result,
        mock_stream_result,
        mock_run_job,
        mock_get_job,
        mock_session,
    ):
        """
        Test run_worker with synchronous handler, ensuring that concurrency behavior
        is respected and that the calls to `get_job`, `run_job`, and `send_result`
        follow expected patterns.

        Args:
            mock_send_result: Mock for send_result function
            mock_stream_result: Mock for stream_result function
            mock_run_job: Mock for run_job function
            mock_get_job: Mock for get_job function
            mock_session: Mock for AsyncClientSession
        """
        # Define the mock behaviors
        mock_get_job.return_value = [{"id": "123", "input": {"number": 1}}]
        mock_run_job.return_value = {"output": {"result": "odd"}}

        # Set a simple concurrency modifier that doesn't change the concurrency
        def concurrency_modifier(current_concurrency):
            return current_concurrency

        config_with_concurrency = self.config.copy()
        config_with_concurrency["concurrency_modifier"] = concurrency_modifier

        # Call the function
        runpod.serverless.start(config_with_concurrency)

        # Make assertions about the behaviors
        self.assertGreaterEqual(
            mock_get_job.call_count, 1, 
            f"Expected at least one call to get_job, but got {mock_get_job.call_count}"
        )

        self.assertGreaterEqual(
            mock_run_job.call_count, 1,
            f"Expected at least one call to run_job, but got {mock_run_job.call_count}"
        )

        self.assertGreaterEqual(
            mock_send_result.call_count, 1,
            f"Expected at least one call to send_result, but got {mock_send_result.call_count}"
        )

        self.assertFalse(
            mock_stream_result.called,
            "stream_result should not be called in this test case."
        )

        self.assertTrue(
            mock_session.called,
            "Expected the mock_session to be used at least once."
        )

        # Verify each call to send_result
        for call in mock_send_result.call_args_list:
            args, kwargs = call
            # Check if the 'output' key is present instead of 'result'
            if "output" in args[1]:
                self.assertIn(
                    "result", args[1]["output"], 
                    "Expected 'result' to be part of the 'output' dictionary."
                )
            else:
                self.fail("The 'output' key was not found in the arguments for send_result.")

    @patch("runpod.serverless.modules.rp_scale.AsyncClientSession")
    @patch("runpod.serverless.modules.rp_scale.get_job")
    @patch("runpod.serverless.modules.rp_job.run_job")
    @patch("runpod.serverless.modules.rp_job.stream_result")
    @patch("runpod.serverless.modules.rp_job.send_result")
    async def test_run_worker_multi_processing(
        self,
        mock_send_result,
        mock_stream_result,
        mock_run_job,
        mock_get_job,
        mock_session,
    ):
        """
        Test run_worker with multi-processing enabled for both async and generator handlers.
        """

        # Define the mock behaviors
        mock_get_job.return_value = [{"id": "123", "input": {"number": 1}}]
        mock_run_job.return_value = {"output": {"result": "odd"}}

        # Run the worker with the original configuration
        runpod.serverless.start(self.config)

        # Check that `get_job`, `run_job`, and `send_result` were called
        self.assertGreaterEqual(mock_get_job.call_count, 1, "Expected at least one call to get_job.")
        self.assertGreaterEqual(mock_run_job.call_count, 1, "Expected at least one call to run_job.")
        self.assertGreaterEqual(mock_send_result.call_count, 1, "Expected at least one call to send_result.")
        
        # Ensure that `stream_result` was not called during the synchronous handler test
        self.assertFalse(mock_stream_result.called, "Expected stream_result to not be called.")
        
        # Ensure that the mock session was used
        self.assertTrue(mock_session.called, "Expected mock session to be called.")

        # Test generator handler
        generator_config = {"handler": generator_handler, "refresh_worker": True}
        runpod.serverless.start(generator_config)

        # Now `stream_result` should be called for the generator handler
        self.assertTrue(mock_stream_result.called, "Expected stream_result to be called for the generator handler.")

        # Test with limited configuration and patch `_set_config_args`
        with patch("runpod.serverless._set_config_args") as mock_set_config_args:
            limited_config = {
                "handler": Mock(),
                "refresh_worker": True,
                "rp_args": {
                    "rp_debugger": True,
                    "rp_serve_api": None,
                    "rp_api_port": 8000,
                    "rp_api_concurrency": 1,
                    "rp_api_host": "localhost",
                },
            }

            mock_set_config_args.return_value = limited_config
            runpod.serverless.start(limited_config)

            # Verify `_set_config_args` was called with the expected arguments
            self.assertTrue(mock_set_config_args.called, "Expected _set_config_args to be called.")
            print(mock_set_config_args.call_args_list)

    @patch("runpod.serverless.modules.rp_scale.get_job")
    @patch("runpod.serverless.modules.rp_job.run_job")
    async def test_run_worker_multi_processing_scaling_up(
        self, mock_run_job, mock_get_job
    ):
        """
        Test run_worker with multi processing enabled, the scale-up and scale-down
        behavior with concurrency_controller.

        Args:
            mock_send_result (_type_): _description_
            mock_stream_result (_type_): _description_
            mock_run_job (_type_): _description_
            mock_get_job (_type_): _description_
            mock_session (_type_): _description_
        """
        # Define the mock behaviors
        mock_get_job.return_value = [{"id": "123", "input": {"number": 1}}]
        mock_run_job.return_value = {"output": {"result": "odd"}}

        # Include multi-processing inside config
        # Should go from concurrency 1 -> 2 -> 4 -> 8 -> 16 -> 8 -> 4 -> 2 -> 1
        # 1+2+4+8+16+8+4+2+1 -> 46 calls to get_job.
        scale_behavior = {
            "behavior": [
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                True,
                True,
                True,
                True,
            ],
            "counter": 0,
        }

        # Let the test be a long running one so we can capture the scale-up and scale-down.
        config = {
            "handler": MagicMock(),
            "refresh_worker": False,
            "rp_args": {"rp_debugger": True, "rp_log_level": "DEBUG"},
        }

        # Let's mock job_scaler.is_alive so that it returns False
        # when scale_behavior's counter is now 5.
        def mock_is_alive():
            res = scale_behavior["counter"] < 10
            scale_behavior["counter"] += 1
            return res

        with patch(
            "runpod.serverless.modules.rp_scale.JobScaler.is_alive", wraps=mock_is_alive
        ):
            runpod.serverless.start(config)

    # Test with sls-core
    async def test_run_worker_with_sls_core(self):
        """
        Test run_worker with sls-core.
        """
        with patch("runpod.serverless.core.main") as mock_main:
            os.environ["RUNPOD_USE_CORE"] = "true"
            runpod.serverless.start(self.config)
            os.environ.pop("RUNPOD_USE_CORE")

            assert mock_main.called

    @patch("runpod.serverless.signal.signal")
    @patch("runpod.serverless.worker.rp_scale.JobScaler.run")
    def test_start_sets_excepthook(self, _, __):
        runpod.serverless.start({})
        assert sys.excepthook == _handle_uncaught_exception

    @patch("runpod.serverless.signal.signal")
    @patch("runpod.serverless.rp_fastapi.WorkerAPI.start_uvicorn")
    @patch("runpod.serverless._set_config_args")
    def test_start_does_not_set_excepthook(self, mock_set_config_args, _, __):
        mock_set_config_args.return_value = self.config
        self.config.update({"rp_args": {
            "rp_serve_api": True,
            "rp_api_host": "localhost",
            "rp_api_port": 8000,
            "rp_api_concurrency": 1,
        }})

        runpod.serverless.start(self.config)
        assert sys.excepthook != _handle_uncaught_exception



""" Tests for CLI group `exec functions` """

import unittest
from unittest.mock import MagicMock, patch

from runpod.cli.groups.exec import functions


class TestExecFunctions(unittest.TestCase):
    """Tests for CLI group `exec functions`"""

    @patch("runpod.cli.groups.exec.functions.ssh_cmd.SSHConnection")
    def test_python_over_ssh(self, mock_ssh_connection):
        """
        Test `python_over_ssh`
        """
        mock_ssh = MagicMock()
        mock_ssh_connection.return_value = mock_ssh

        pod_id = "pod_id"
        file_name = "file_name"

        functions.python_over_ssh(pod_id, file_name)

        self.assertTrue(functions.python_over_ssh)
        mock_ssh_connection.assert_called_once_with(pod_id)
        mock_ssh.put_file.assert_called_once_with(file_name, f"/root/{file_name}")
        mock_ssh.run_commands.assert_called_once_with([f"python3.10 /root/{file_name}"])
        mock_ssh.close.assert_called_once_with()



""" Tests for rp_local.py """

from unittest import IsolatedAsyncioTestCase
from unittest.mock import mock_open, patch

from runpod.serverless.modules import rp_local


class TestRunLocal(IsolatedAsyncioTestCase):
    """Tests for run_local function"""

    @patch(
        "runpod.serverless.modules.rp_local.run_job", return_value={"result": "success"}
    )
    @patch("builtins.open", new_callable=mock_open, read_data='{"input": "test"}')
    async def test_run_local_with_test_input(self, mock_file, mock_run):
        """
        Test run_local function with test_input in rp_args
        """
        config = {
            "handler": "handler",
            "rp_args": {
                "test_input": {"input": "test", "id": "test_id"},
                "test_output": {"result": "success"},
            },
        }
        with self.assertRaises(SystemExit) as sys_exit:
            await rp_local.run_local(config)
            self.assertEqual(sys_exit.exception.code, 0)

        config["rp_args"]["test_output"] = {"result": "fail"}
        with self.assertRaises(SystemExit) as sys_exit:
            await rp_local.run_local(config)
            self.assertEqual(sys_exit.exception.code, 1)

        assert mock_file.called is False
        assert mock_run.called

    @patch("runpod.serverless.modules.rp_local.run_job", return_value={})
    @patch("builtins.open", new_callable=mock_open, read_data='{"input": "test"}')
    async def test_run_local_with_test_input_json(self, mock_file, mock_run):
        """
        Test run_local function with test_input.json
        """
        config = {"handler": "handler", "rp_args": {}}
        with patch("os.path.exists", return_value=True):
            with self.assertRaises(SystemExit) as sys_exit:
                await rp_local.run_local(config)
            self.assertEqual(sys_exit.exception.code, 0)

        assert mock_file.called
        assert mock_run.called

    @patch(
        "runpod.serverless.modules.rp_local.run_job",
        return_value={"error": "test_error"},
    )
    @patch("builtins.open", new_callable=mock_open, read_data='{"input": "test"}')
    async def test_run_local_with_error(self, mock_file, mock_run):
        """
        Test run_local function when run_job returns an error
        """
        config = {
            "handler": "handler",
            "rp_args": {"test_input": {"input": "test", "id": "test_id"}},
        }
        with self.assertRaises(SystemExit) as sys_exit:
            await rp_local.run_local(config)
        self.assertEqual(sys_exit.exception.code, 1)

        assert mock_file.called is False
        assert mock_run.called

    async def test_run_local_without_test_input_json(self):
        """
        Test run_local function without test_input.json
        """
        config = {"handler": "handler", "rp_args": {}}
        with patch("os.path.exists", return_value=False):
            with self.assertRaises(SystemExit) as sys_exit:
                await rp_local.run_local(config)
            self.assertEqual(sys_exit.exception.code, 1)

    @patch("runpod.serverless.modules.rp_local.run_job", return_value={})
    @patch("builtins.open", new_callable=mock_open, read_data='{"not_input": "test"}')
    async def test_run_local_without_input(self, mock_file, mock_run):
        """
        Test run_local function without input in test_input.json
        """
        config = {"handler": "handler", "rp_args": {}}
        with patch("os.path.exists", return_value=True):
            with self.assertRaises(SystemExit) as sys_exit:
                await rp_local.run_local(config)
            self.assertEqual(sys_exit.exception.code, 1)

        assert mock_file.called
        assert mock_run.called is False



"""
RunPod | API Wrapper | Mutations | Pods
"""

# pylint: disable=too-many-arguments, too-many-locals, too-many-branches

from typing import List, Optional


def generate_pod_deployment_mutation(
    name: str,
    image_name: str,
    gpu_type_id: str,
    cloud_type: str = "ALL",
    support_public_ip: bool = True,
    start_ssh: bool = True,
    data_center_id=None,
    country_code=None,
    gpu_count=None,
    volume_in_gb=None,
    container_disk_in_gb=None,
    min_vcpu_count=None,
    min_memory_in_gb=None,
    docker_args=None,
    ports=None,
    volume_mount_path=None,
    env: dict = None,
    template_id=None,
    network_volume_id=None,
    allowed_cuda_versions: Optional[List[str]] = None,
    min_download=None,
    min_upload=None,
):
    """
    Generates a mutation to deploy a pod on demand.
    """
    input_fields = []

    # ------------------------------ Required Fields ----------------------------- #
    input_fields.append(f'name: "{name}"')
    input_fields.append(f'imageName: "{image_name}"')
    input_fields.append(f'gpuTypeId: "{gpu_type_id}"')

    # ------------------------------ Default Fields ------------------------------ #
    input_fields.append(f"cloudType: {cloud_type}")

    if start_ssh:
        input_fields.append("startSsh: true")

    if support_public_ip:
        input_fields.append("supportPublicIp: true")
    else:
        input_fields.append("supportPublicIp: false")

    # ------------------------------ Optional Fields ----------------------------- #
    if data_center_id is not None:
        input_fields.append(f'dataCenterId: "{data_center_id}"')
    if country_code is not None:
        input_fields.append(f'countryCode: "{country_code}"')
    if gpu_count is not None:
        input_fields.append(f"gpuCount: {gpu_count}")
    if volume_in_gb is not None:
        input_fields.append(f"volumeInGb: {volume_in_gb}")
    if container_disk_in_gb is not None:
        input_fields.append(f"containerDiskInGb: {container_disk_in_gb}")
    if min_vcpu_count is not None:
        input_fields.append(f"minVcpuCount: {min_vcpu_count}")
    if min_memory_in_gb is not None:
        input_fields.append(f"minMemoryInGb: {min_memory_in_gb}")
    if docker_args is not None:
        input_fields.append(f'dockerArgs: "{docker_args}"')
    if ports is not None:
        ports = ports.replace(" ", "")
        input_fields.append(f'ports: "{ports}"')
    if volume_mount_path is not None:
        input_fields.append(f'volumeMountPath: "{volume_mount_path}"')
    if env is not None:
        env_string = ", ".join(
            [f'{{ key: "{key}", value: "{value}" }}' for key, value in env.items()]
        )
        input_fields.append(f"env: [{env_string}]")
    if template_id is not None:
        input_fields.append(f'templateId: "{template_id}"')

    if network_volume_id is not None:
        input_fields.append(f'networkVolumeId: "{network_volume_id}"')

    if allowed_cuda_versions is not None:
        allowed_cuda_versions_string = ", ".join(
            [f'"{version}"' for version in allowed_cuda_versions]
        )
        input_fields.append(f"allowedCudaVersions: [{allowed_cuda_versions_string}]")

    if min_download is not None:
        input_fields.append(f'minDownload: {min_download}')

    if min_upload is not None:
        input_fields.append(f'minUpload: {min_upload}')

    # Format input fields
    input_string = ", ".join(input_fields)
    return f"""
    mutation {{
      podFindAndDeployOnDemand(
        input: {{
          {input_string}
        }}
      ) {{
        id
        desiredStatus
        imageName
        env
        machineId
        machine {{
          podHostId
        }}
      }}
    }}
    """


def generate_pod_stop_mutation(pod_id: str) -> str:
    """
    Generates a mutation to stop a pod.
    """
    return f"""
    mutation {{
        podStop(input: {{ podId: "{pod_id}" }}) {{
            id
            desiredStatus
        }}
    }}
    """


def generate_pod_resume_mutation(pod_id: str, gpu_count: int) -> str:
    """
    Generates a mutation to resume a pod.
    """
    return f"""
    mutation {{
        podResume(input: {{ podId: "{pod_id}", gpuCount: {gpu_count} }}) {{
            id
            desiredStatus
            imageName
            env
            machineId
            machine {{
                podHostId
            }}
        }}
    }}
    """


def generate_pod_terminate_mutation(pod_id: str) -> str:
    """
    Generates a mutation to terminate a pod.
    """
    return f"""
    mutation {{
        podTerminate(input: {{ podId: "{pod_id}" }})
    }}
    """



""" Reads the .runpodignore file and returns a list of files to ignore. """

import fnmatch
import os

EXCLUDE_PATTERNS = [
    "__pycache__/",
    "*.pyc",
    ".*.swp",
    ".git/",
    "*.tmp",
    "*.log",
]


def get_ignore_list():
    """Reads the .runpodignore file and returns a list of files to ignore."""
    ignore_list = EXCLUDE_PATTERNS.copy()
    ignore_file = os.path.join(os.getcwd(), ".runpodignore")

    if not os.path.isfile(ignore_file):
        return ignore_list

    with open(ignore_file, "r", encoding="UTF-8") as ignore_file_handle:
        for line in ignore_file_handle:
            stripped_line = line.strip()
            if stripped_line and not stripped_line.startswith("#"):
                ignore_list.append(stripped_line)

    return ignore_list


def should_ignore(file_path, ignore_list=None):
    """Returns True if the file should be ignored, False otherwise."""
    if ignore_list is None:
        ignore_list = get_ignore_list()

    relative_path = os.path.relpath(file_path, os.getcwd())

    for pattern in ignore_list:
        if pattern.startswith("/"):
            pattern = pattern[1:]

        if pattern.endswith("/"):
            pattern += "*"

        if fnmatch.fnmatch(relative_path, pattern):
            return True

    return False



"""
RunPod | API Wrapper | CTL Commands
"""

# pylint: disable=too-many-arguments,too-many-locals

from typing import Optional

from .graphql import run_graphql_query
from .mutations import container_register_auth as container_register_auth_mutations
from .mutations import endpoints as endpoint_mutations
from .mutations import pods as pod_mutations
from .mutations import templates as template_mutations
from .mutations import user as user_mutations
from .queries import endpoints as endpoint_queries
from .queries import gpus
from .queries import pods as pod_queries
from .queries import user as user_queries


def get_user() -> dict:
    """
    Get the current user
    """
    raw_response = run_graphql_query(user_queries.QUERY_USER)
    cleaned_return = raw_response["data"]["myself"]
    return cleaned_return


def update_user_settings(pubkey: str) -> dict:
    """
    Update the current user

    :param pubkey: the public key of the user
    """
    raw_response = run_graphql_query(user_mutations.generate_user_mutation(pubkey))
    cleaned_return = raw_response["data"]["updateUserSettings"]
    return cleaned_return


def get_gpus() -> dict:
    """
    Get all GPU types
    """
    raw_response = run_graphql_query(gpus.QUERY_GPU_TYPES)
    cleaned_return = raw_response["data"]["gpuTypes"]
    return cleaned_return


def get_gpu(gpu_id: str, gpu_quantity: int = 1):
    """
    Get a specific GPU type

    :param gpu_id: the id of the gpu
    :param gpu_quantity: how many of the gpu should be returned
    """
    raw_response = run_graphql_query(gpus.generate_gpu_query(gpu_id, gpu_quantity))

    cleaned_return = raw_response["data"]["gpuTypes"]

    if len(cleaned_return) < 1:
        raise ValueError(
            "No GPU found with the specified ID, "
            "run runpod.get_gpus() to get a list of all GPUs"
        )

    return cleaned_return[0]


def get_pods() -> dict:
    """
    Get all pods
    """
    raw_return = run_graphql_query(pod_queries.QUERY_POD)
    cleaned_return = raw_return["data"]["myself"]["pods"]
    return cleaned_return


def get_pod(pod_id: str):
    """
    Get a specific pod

    :param pod_id: the id of the pod
    """
    raw_response = run_graphql_query(pod_queries.generate_pod_query(pod_id))
    return raw_response["data"]["pod"]


def create_pod(
    name: str,
    image_name: str,
    gpu_type_id: str,
    cloud_type: str = "ALL",
    support_public_ip: bool = True,
    start_ssh: bool = True,
    data_center_id: Optional[str] = None,
    country_code: Optional[str] = None,
    gpu_count: int = 1,
    volume_in_gb: int = 0,
    container_disk_in_gb: Optional[int] = None,
    min_vcpu_count: int = 1,
    min_memory_in_gb: int = 1,
    docker_args: str = "",
    ports: Optional[str] = None,
    volume_mount_path: str = "/runpod-volume",
    env: Optional[dict] = None,
    template_id: Optional[str] = None,
    network_volume_id: Optional[str] = None,
    allowed_cuda_versions: Optional[list] = None,
    min_download = None,
    min_upload = None,
) -> dict:
    """
    Create a pod

    :param name: the name of the pod
    :param image_name: the name of the docker image to be used by the pod
    :param gpu_type_id: the gpu type wanted by the pod (retrievable by get_gpus)
    :param cloud_type: if secure cloud, community cloud or all is wanted
    :param data_center_id: the id of the data center
    :param country_code: the code for country to start the pod in
    :param gpu_count: how many gpus should be attached to the pod
    :param volume_in_gb: how big should the pod volume be
    :param ports: the ports to open in the pod, example format - "8888/http,666/tcp"
    :param volume_mount_path: where to mount the volume?
    :param env: the environment variables to inject into the pod,
                for example {EXAMPLE_VAR:"example_value", EXAMPLE_VAR2:"example_value 2"}, will
                inject EXAMPLE_VAR and EXAMPLE_VAR2 into the pod with the mentioned values
    :param template_id: the id of the template to use for the pod
    :param min_download: minimum download speed in Mbps
    :param min_upload: minimum upload speed in Mbps
    :example:

    >>> pod_id = runpod.create_pod("test", "runpod/stack", "NVIDIA GeForce RTX 3070")
    """
    # Input Validation
    get_gpu(gpu_type_id)  # Check if GPU exists, will raise ValueError if not.
    if cloud_type not in ["ALL", "COMMUNITY", "SECURE"]:
        raise ValueError("cloud_type must be one of ALL, COMMUNITY or SECURE")

    if network_volume_id and data_center_id is None:
        user_info = get_user()
        for network_volume in user_info["networkVolumes"]:
            if network_volume["id"] == network_volume_id:
                data_center_id = network_volume["dataCenterId"]
                break

    if container_disk_in_gb is None and template_id is None:
        container_disk_in_gb = 10

    raw_response = run_graphql_query(
        pod_mutations.generate_pod_deployment_mutation(
            name,
            image_name,
            gpu_type_id,
            cloud_type,
            support_public_ip,
            start_ssh,
            data_center_id,
            country_code,
            gpu_count,
            volume_in_gb,
            container_disk_in_gb,
            min_vcpu_count,
            min_memory_in_gb,
            docker_args,
            ports,
            volume_mount_path,
            env,
            template_id,
            network_volume_id,
            allowed_cuda_versions,
            min_download,
            min_upload,
        )
    )

    cleaned_response = raw_response["data"]["podFindAndDeployOnDemand"]
    return cleaned_response


def stop_pod(pod_id: str):
    """
    Stop a pod

    :param pod_id: the id of the pod

    :example:

    >>> pod_id = runpod.create_pod("test", "runpod/stack", "NVIDIA GeForce RTX 3070")
    >>> runpod.stop_pod(pod_id)
    """
    raw_response = run_graphql_query(pod_mutations.generate_pod_stop_mutation(pod_id))

    cleaned_response = raw_response["data"]["podStop"]
    return cleaned_response


def resume_pod(pod_id: str, gpu_count: int):
    """
    Resume a pod

    :param pod_id: the id of the pod
    :param gpu_count: the number of GPUs to attach to the pod

    :example:

    >>> pod_id = runpod.create_pod("test", "runpod/stack", "NVIDIA GeForce RTX 3070")
    >>> runpod.stop_pod(pod_id)
    >>> runpod.resume_pod(pod_id)
    """
    raw_response = run_graphql_query(
        pod_mutations.generate_pod_resume_mutation(pod_id, gpu_count)
    )

    cleaned_response = raw_response["data"]["podResume"]
    return cleaned_response


def terminate_pod(pod_id: str):
    """
    Terminate a pod

    :param pod_id: the id of the pod

    :example:

    >>> pod_id = runpod.create_pod("test", "runpod/stack", "NVIDIA GeForce RTX 3070")
    >>> runpod.terminate_pod(pod_id)
    """
    run_graphql_query(pod_mutations.generate_pod_terminate_mutation(pod_id))


def create_template(
    name: str,
    image_name: str,
    docker_start_cmd: str = None,
    container_disk_in_gb: int = 10,
    volume_in_gb: int = None,
    volume_mount_path: str = None,
    ports: str = None,
    env: dict = None,
    is_serverless: bool = False,
    registry_auth_id: str = None,
):
    """
    Create a template

    :param name: the name of the template
    :param image_name: the name of the docker image to be used by the template
    :param docker_start_cmd: the command to start the docker container with
    :param container_disk_in_gb: how big should the container disk be
    :param volume_in_gb: how big should the volume be
    :param ports: the ports to open in the pod, example format - "8888/http,666/tcp"
    :param volume_mount_path: where to mount the volume?
    :param env: the environment variables to inject into the pod,
                for example {EXAMPLE_VAR:"example_value", EXAMPLE_VAR2:"example_value 2"}, will
                inject EXAMPLE_VAR and EXAMPLE_VAR2 into the pod with the mentioned values
    :param is_serverless: is the template serverless?
    :param registry_auth_id: the id of the registry auth

    :example:

    >>> template_id = runpod.create_template("test", "runpod/stack", "python3 main.py")
    """
    raw_response = run_graphql_query(
        template_mutations.generate_pod_template(
            name=name,
            image_name=image_name,
            docker_start_cmd=docker_start_cmd,
            container_disk_in_gb=container_disk_in_gb,
            volume_in_gb=volume_in_gb,
            volume_mount_path=volume_mount_path,
            ports=ports,
            env=env,
            is_serverless=is_serverless,
            registry_auth_id=registry_auth_id,
        )
    )

    return raw_response["data"]["saveTemplate"]


def get_endpoints() -> dict:
    """
    Get all endpoints
    """
    raw_return = run_graphql_query(endpoint_queries.QUERY_ENDPOINT)
    cleaned_return = raw_return["data"]["myself"]["endpoints"]
    return cleaned_return


def create_endpoint(
    name: str,
    template_id: str,
    gpu_ids: str = "AMPERE_16",
    network_volume_id: str = None,
    locations: str = None,
    idle_timeout: int = 5,
    scaler_type: str = "QUEUE_DELAY",
    scaler_value: int = 4,
    workers_min: int = 0,
    workers_max: int = 3,
    flashboot=False,
    allowed_cuda_versions: str = "12.1,12.2,12.3,12.4,12.5",
    gpu_count: int = 1,
):
    """
    Create an endpoint

    :param allowed_cuda_versions: Comma-separated string of allowed CUDA versions (e.g., "12.4,12.5").
    """
    raw_response = run_graphql_query(
        endpoint_mutations.generate_endpoint_mutation(
            name,
            template_id,
            gpu_ids,
            network_volume_id,
            locations,
            idle_timeout,
            scaler_type,
            scaler_value,
            workers_min,
            workers_max,
            flashboot,
            allowed_cuda_versions,
            gpu_count
        )
    )

    return raw_response["data"]["saveEndpoint"]


def update_endpoint_template(endpoint_id: str, template_id: str):
    """
    Update an endpoint template

    :param endpoint_id: the id of the endpoint
    :param template_id: the id of the template to use for the endpoint

    :example:

    >>> endpoint_id = runpod.update_endpoint_template("test", "template_id")
    """
    raw_response = run_graphql_query(
        endpoint_mutations.update_endpoint_template_mutation(endpoint_id, template_id)
    )

    return raw_response["data"]["updateEndpointTemplate"]


def create_container_registry_auth(name: str, username: str, password: str):
    """
    Create a container registry authentication.

    Args:
        name (str): The name of the container registry.
        username (str): The username for authentication.
        password (str): The password for authentication.

    Returns:
        dict: The response data containing the saved container registry authentication.
    """
    raw_response = run_graphql_query(
        container_register_auth_mutations.generate_container_registry_auth(
            name, username, password
        )
    )
    return raw_response["data"]["saveRegistryAuth"]


def update_container_registry_auth(registry_auth_id: str, username: str, password: str):
    """
    Update a container registry authentication.

    Args:
        registry_auth_id (str): The id of the container registry authentication
        username (str): The username for authentication.
        password (str): The password for authentication.

    Returns:
        dict: The response data containing the updated container registry authentication.
    """
    raw_response = run_graphql_query(
        container_register_auth_mutations.update_container_registry_auth(
            registry_auth_id, username, password
        )
    )
    return raw_response["data"]["updateRegistryAuth"]


def delete_container_registry_auth(registry_auth_id: str):
    """
    Delete a container registry authentication.

    Args:
        registry_auth_id (str): The id of the container registry authentication
    """
    raw_response = run_graphql_query(
        container_register_auth_mutations.delete_container_registry_auth(
            registry_auth_id
        )
    )
    return raw_response["data"]["deleteRegistryAuth"]



"""
This is the starting point for your serverless API.
It can be named anything you want, but needs to have the following:

- import runpod
- start function

To return an error, return a dictionary with the key "error" and the value being the error message.
"""

import runpod  # Required


def is_even(job):
    """
    Example function that returns True if the input is even, False otherwise.

    "job_input" will contain the input that was passed to the API along with some other metadata.
    The structure will look like this:
    {
        "id": "some-id",
        "input": {"number": 2}
    }

    Whatever is returned from this function will be returned to the user as the output.
    """

    job_input = job["input"]
    the_number = job_input["number"]

    if not isinstance(the_number, int):
        return {"error": "Silly human, you need to pass an integer."}

    if the_number % 2 == 0:
        return True

    return False


if __name__ == "__main__":
    runpod.serverless.start({"handler": is_even})



"""
This module defines the Heartbeat class.
The heartbeat is responsible for sending periodic pings to the Runpod server.
"""

import os
import threading
import time

import requests
from urllib3.util.retry import Retry

from runpod.http_client import SyncClientSession
from runpod.serverless.modules.rp_logger import RunPodLogger
from runpod.serverless.modules.worker_state import WORKER_ID, JobsProgress
from runpod.version import __version__ as runpod_version

log = RunPodLogger()
jobs = JobsProgress()  # Contains the list of jobs that are currently running.


class Heartbeat:
    """Sends heartbeats to the Runpod server."""

    _thread_started = False

    def __init__(self, pool_connections=10, retries=3) -> None:
        """
        Initializes the Heartbeat class.
        """
        self.PING_URL = os.environ.get("RUNPOD_WEBHOOK_PING", "PING_NOT_SET")
        self.PING_URL = self.PING_URL.replace("$RUNPOD_POD_ID", WORKER_ID)
        self.PING_INTERVAL = int(os.environ.get("RUNPOD_PING_INTERVAL", 10000)) // 1000

        self._session = SyncClientSession()
        self._session.headers.update(
            {"Authorization": f"{os.environ.get('RUNPOD_AI_API_KEY')}"}
        )

        retry_strategy = Retry(
            total=retries,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["GET"],
            backoff_factor=1,
        )

        adapter = requests.adapters.HTTPAdapter(
            pool_connections=pool_connections,
            pool_maxsize=pool_connections,
            max_retries=retry_strategy,
        )
        self._session.mount("http://", adapter)
        self._session.mount("https://", adapter)

    def start_ping(self, test=False):
        """
        Sends heartbeat pings to the Runpod server.
        """
        if not os.environ.get("RUNPOD_AI_API_KEY"):
            log.debug("Not deployed on RunPod serverless, pings will not be sent.")
            return

        if not os.environ.get("RUNPOD_POD_ID"):
            log.info("Not running on RunPod, pings will not be sent.")
            return

        if (not self.PING_URL) or self.PING_URL == "PING_NOT_SET":
            log.error("Ping URL not set, cannot start ping.")
            return

        if not Heartbeat._thread_started:
            threading.Thread(target=self.ping_loop, daemon=True, args=(test,)).start()
            Heartbeat._thread_started = True

    def ping_loop(self, test=False):
        """
        Sends heartbeat pings to the Runpod server.
        """
        while True:
            self._send_ping()
            time.sleep(self.PING_INTERVAL)

            if test:
                return

    def _send_ping(self):
        """
        Sends a heartbeat to the Runpod server.
        """
        job_ids = jobs.get_job_list()
        ping_params = {"job_id": job_ids, "runpod_version": runpod_version}

        try:
            result = self._session.get(
                self.PING_URL, params=ping_params, timeout=self.PING_INTERVAL * 2
            )

            log.debug(
                f"Heartbeat Sent | URL: {result.url} | Status: {result.status_code}"
            )

        except requests.RequestException as err:
            log.error(f"Ping Request Error: {err}, attempting to restart ping.")



""" GraphQL queries for endpoints. """

QUERY_ENDPOINT = """
query Query {
  myself {
    endpoints {
      aiKey
      gpuIds
      id
      idleTimeout
      name
      networkVolumeId
      locations
      scalerType
      scalerValue
      templateId
      type
      userId
      version
      workersMax
      workersMin
      workersStandby
      gpuCount
      env {
        key
        value
      }
      createdAt
      networkVolume {
        id
        dataCenterId
      }
    }
  }
}
"""



"""
HTTP Client abstractions
"""

import os

import requests
from aiohttp import ClientSession, ClientTimeout, TCPConnector, ClientResponseError

from .cli.groups.config.functions import get_credentials
from .user_agent import USER_AGENT


class TooManyRequests(ClientResponseError):
    pass


def get_auth_header():
    """
    Produce a header dict with the `Authorization` key derived from
    credentials.get("api_key") OR os.getenv('RUNPOD_AI_API_KEY')
    """
    if credentials := get_credentials():
        auth = credentials.get("api_key", "")
    else:
        auth = os.getenv("RUNPOD_AI_API_KEY", "")

    return {
        "Content-Type": "application/json",
        "Authorization": auth,
        "User-Agent": USER_AGENT,
    }


def AsyncClientSession(*args, **kwargs):  # pylint: disable=invalid-name
    """
    Deprecation from aiohttp.ClientSession forbids inheritance.
    This is now a factory method
    """
    return ClientSession(
        connector=TCPConnector(limit=0),
        headers=get_auth_header(),
        timeout=ClientTimeout(600, ceil_threshold=400),
        *args,
        **kwargs,
    )


class SyncClientSession(requests.Session):
    """
    Inherits requests.Session to override `request()` method for tracing
    """
    pass


"""
RunPod | CLI | Project | Commands
"""

import os
import sys

import click
from InquirerPy import prompt as cli_select

from runpod import get_user

from .functions import create_new_project, create_project_endpoint, start_project
from .helpers import validate_project_name


@click.group("project", help="Create, deploy, and manage RunPod projects.")
def project_cli():
    """Create and deploy projects on RunPod."""


# -------------------------------- New Project ------------------------------- #
@project_cli.command("new")
@click.option(
    "--name", "-n", "project_name", type=str, default=None, help="The project name."
)
@click.option(
    "--type",
    "-t",
    "model_type",
    type=click.Choice(["llama2"], case_sensitive=False),
    default=None,
    help="The type of Hugging Face model.",
)
@click.option(
    "--model",
    "-m",
    "model_name",
    type=str,
    default=None,
    help="The name of the Hugging Face model. (e.g. meta-llama/Llama-2-7b)",
)
@click.option("--init", "-i", "init_current_dir", is_flag=True, default=False)
def new_project_wizard(project_name, model_type, model_name, init_current_dir):
    """Create a new project."""
    network_volumes = get_user()["networkVolumes"]
    if len(network_volumes) == 0:
        click.echo("You do not have any network volumes.")
        click.echo(
            "Please create a network volume (https://runpod.io/console/user/storage) and try again."
        )  # pylint: disable=line-too-long
        sys.exit(1)

    click.echo("Creating a new project...")

    if init_current_dir:
        project_name = os.path.basename(os.getcwd())

    if project_name is None:
        project_name = click.prompt("   > Enter the project name", type=str)

    validate_project_name(project_name)

    def print_net_vol(vol):
        return {
            "name": f"{vol['id']}: {vol['name']} ({vol['size']} GB, {vol['dataCenterId']})",
            "value": vol["id"],
        }

    network_volumes = list(map(print_net_vol, network_volumes))
    questions = [
        {
            "type": "rawlist",
            "name": "volume-id",
            "qmark": "",
            "amark": "",
            "message": "   > Select a Network Volume:",
            "choices": network_volumes,
        }
    ]
    runpod_volume_id = cli_select(questions)["volume-id"]

    cuda_version = click.prompt(
        "   > Select a CUDA version, or press enter to use the default",
        type=click.Choice(["11.1.1", "11.8.0", "12.1.0"], case_sensitive=False),
        default="11.8.0",
    )

    python_version = click.prompt(
        "   > Select a Python version, or press enter to use the default",
        type=click.Choice(["3.8", "3.9", "3.10", "3.11"], case_sensitive=False),
        default="3.10",
    )

    click.echo("")
    click.echo("Project Summary:")

    click.echo(f"   - Project Name: {project_name}")
    click.echo(f"   - RunPod Network Storage ID: {runpod_volume_id}")
    click.echo(f"   - CUDA Version: {cuda_version}")
    click.echo(f"   - Python Version: {python_version}")

    click.echo("")
    click.echo("The project will be created in the current directory.")
    click.confirm("Do you want to continue?", abort=True)

    create_new_project(
        project_name,
        runpod_volume_id,
        cuda_version,
        python_version,
        model_type,
        model_name,
        init_current_dir,
    )  # pylint: disable=too-many-arguments

    click.echo(f"Project {project_name} created successfully!")
    click.echo("")
    click.echo(
        "From your project root run `runpod project start` to start a development pod."
    )


# ------------------------------- Start Project ------------------------------ #
@project_cli.command("start")
def start_project_pod():
    """
    Start the project development pod from runpod.toml
    """
    click.echo(
        "Starting the project will create a development pod on RunPod, if one does not already exist."
    )  # pylint: disable=line-too-long
    click.echo(
        "    - You will be charged based on the GPU type specified in runpod.toml."
    )
    click.echo("    - When you are finished close the connection with CTRL+C.")
    click.echo("")
    click.confirm("Do you want to continue?", abort=True)

    click.echo("Starting project development pod...")

    start_project()


# ------------------------------ Deploy Project ------------------------------ #
@project_cli.command("deploy")
def deploy_project():
    """Deploy the project to RunPod."""
    click.echo("Deploying project...")

    endpoint_id = create_project_endpoint()

    click.echo(f"Project deployed successfully! Endpoint ID: {endpoint_id}")
    click.echo("The following urls are available:")
    click.echo(f"    - https://api.runpod.ai/v2/{endpoint_id}/runsync")
    click.echo(f"    - https://api.runpod.ai/v2/{endpoint_id}/run")
    click.echo(f"    - https://api.runpod.ai/v2/{endpoint_id}/health")



""" User-Agent for RunPod-Python-SDK """

import os
import platform

from runpod.version import __version__ as runpod_version


def construct_user_agent():
    """Constructs the User-Agent string for the RunPod-Python-SDK

    Example:
        RunPod-Python-SDK/0.1.0 (Linux 5.4.0-54-generic; x86_64) Language/Python 3.8.5
    """
    os_info = f"{platform.system()} {platform.release()}; {platform.machine()}"
    python_version = platform.python_version()
    integration_method = os.getenv("RUNPOD_UA_INTEGRATION")

    ua_components = [
        f"RunPod-Python-SDK/{runpod_version}",
        f"({os_info})",
        f"Language/Python {python_version}",
    ]

    if integration_method:
        ua_components.append(f"Integration/{integration_method}")

    user_agent = " ".join(ua_components)
    return user_agent


USER_AGENT = construct_user_agent()



""" Unit tests for the runpodignore module in the cli.utils package. """

# pylint: disable=duplicate-code

import os
import unittest
from unittest.mock import mock_open, patch

from runpod.cli.utils.rp_runpodignore import get_ignore_list, should_ignore


class TestGetIgnoreList(unittest.TestCase):
    """Unit tests for the get_ignore_list function."""

    @patch("os.path.isfile", return_value=False)
    def test_no_ignore_file(self, mock_isfile):
        """Test that the default ignore list is returned when no ignore file is present."""
        result = get_ignore_list()
        self.assertEqual(
            result, ["__pycache__/", "*.pyc", ".*.swp", ".git/", "*.tmp", "*.log"]
        )
        assert mock_isfile.called

    @patch("os.path.isfile", return_value=True)
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="*.hello\n# This is a comment\n\n.world\n",
    )  # pylint: disable=line-too-long
    def test_with_ignore_file(self, mock_file, mock_isfile):
        """Test that the default ignore list is returned when no ignore file is present."""
        result = get_ignore_list()
        expected_patterns = [
            "__pycache__/",
            "*.pyc",
            ".*.swp",
            ".git/",
            "*.tmp",
            "*.log",
            "*.hello",
            ".world",
        ]
        self.assertEqual(result, expected_patterns)
        assert mock_file.called
        assert mock_isfile.called

    @patch("os.path.isfile", return_value=True)
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data="# Only comments and empty lines\n\n\n",
    )  # pylint: disable=line-too-long
    def test_empty_ignore_file(self, mock_file, mock_isfile):
        """Test that the default ignore list is returned when no ignore file is present."""
        result = get_ignore_list()
        self.assertEqual(
            result, ["__pycache__/", "*.pyc", ".*.swp", ".git/", "*.tmp", "*.log"]
        )
        assert mock_file.called
        assert mock_isfile.called


class TestShouldIgnoreFunction(unittest.TestCase):
    """Unit tests for the should_ignore function."""

    def test_should_ignore_with_relative_path(self):
        """Test that the should_ignore function returns True when the file should be ignored."""
        patterns = ["/test/"]
        file_path = os.path.join(os.getcwd(), "test", "example.txt")
        self.assertTrue(should_ignore(file_path, patterns))

    def test_should_ignore_with_absolute_path(self):
        """Test that the should_ignore function returns True when the file should be ignored."""
        patterns = ["*.txt"]
        file_path = os.path.join(os.getcwd(), "test", "example.txt")
        self.assertTrue(should_ignore(file_path, patterns))

    def test_should_not_ignore(self):
        """Test that the should_ignore
        Function returns False when the file should not be ignored
        """
        patterns = ["*.md"]
        file_path = os.path.join(os.getcwd(), "test", "example.txt")
        self.assertFalse(should_ignore(file_path, patterns))

    @patch("runpod.cli.utils.rp_runpodignore.get_ignore_list")
    def test_should_ignore_with_mocked_ignore_list(self, mock_get_ignore_list):
        """Test that the should_ignore function returns True when the file should be ignored."""
        mock_get_ignore_list.return_value = ["*.txt"]
        file_path = "example.txt"
        self.assertTrue(should_ignore(file_path))

    @patch("runpod.cli.utils.rp_runpodignore.get_ignore_list")
    def test_should_not_ignore_with_mocked_ignore_list(self, mock_get_ignore_list):
        """Test that the should_ignore
        Function returns False when the file should not be ignored
        """
        mock_get_ignore_list.return_value = ["*.md"]
        file_path = "example.txt"
        self.assertFalse(should_ignore(file_path))

    @patch("runpod.cli.utils.rp_runpodignore.os.getcwd")
    def test_should_ignore_with_mocked_cwd(self, mock_getcwd):
        """Test that the should_ignore function returns True when the file should be ignored."""
        mock_getcwd.return_value = "/fake/directory"
        patterns = ["/test/"]
        file_path = "/fake/directory/test/example.txt"
        self.assertTrue(should_ignore(file_path, patterns))



""" Used to launch the FastAPI web server when worker is running in API mode. """

import os
import threading
import uuid
from dataclasses import dataclass
from typing import Any, Dict, Optional, Union

import requests
import uvicorn
from fastapi import APIRouter, FastAPI
from fastapi.encoders import jsonable_encoder
from fastapi.responses import RedirectResponse

from ...http_client import SyncClientSession
from ...version import __version__ as runpod_version
from .rp_handler import is_generator
from .rp_job import run_job, run_job_generator
from .rp_ping import Heartbeat
from .worker_state import Job, JobsProgress

RUNPOD_ENDPOINT_ID = os.environ.get("RUNPOD_ENDPOINT_ID", None)

TITLE = "RunPod | Development Worker API"

DESCRIPTION = """
The Development Worker API facilitates testing and debugging of your RunPod workers.
It offers a sandbox environment for executing code and simulating interactions with your worker, ensuring your applications can seamlessly transition to production on RunPod serverless platform.
Use this API for comprehensive testing of request submissions and result retrieval, mimicking the behavior of RunPod's operational environment.
---
*Note: This API serves as a local testing tool and will not be utilized once your worker is operational on the RunPod platform.*
"""

# Add CLI tool suggestion if RUNPOD_PROJECT_ID is not set.
if os.environ.get("RUNPOD_PROJECT_ID", None) is None:
    DESCRIPTION += """

    ℹ️ | Consider developing with our CLI tool to streamline your worker development process.

    >_  wget -qO- cli.runpod.net | sudo bash
    >_  runpodctl project create
    """

RUN_DESCRIPTION = """
Initiates processing jobs, returning a unique job ID.

**Parameters:**
- **input** (string): The data to be processed by the worker. This could be a string, JSON object, etc., depending on the worker's requirements.
- **webhook** (string, optional): A callback URL for result notification upon completion. If specified, the server will send a POST request to this URL with the job's result once it's available.

**Returns:**
- **job_id** (string): A unique identifier for the job, used with the `/stream` and `/status` endpoints for monitoring progress and checking job status.
"""

RUNSYNC_DESCRIPTION = """
Executes processing jobs synchronously, returning the job's output directly.

This endpoint is ideal for tasks where immediate result retrieval is necessary,
streamlining the execution process by eliminating the need for subsequent
status or result checks.

**Parameters:**
- **input** (string): The data to be processed by the worker. This should be in a format that the worker can understand (e.g., JSON, text, etc.).
- **webhook** (string, optional): A callback URL to which the result will be posted. While direct result retrieval is the primary operation mode for this endpoint, specifying a webhook allows for asynchronous result notification if needed.

**Returns:**
- **output** (Any): The direct output from the processing job, formatted according to the job's nature and the expected response structure. This could be a JSON object, plain text, or any data structure depending on the processing logic.
"""

STREAM_DESCRIPTION = """
Continuously aggregates the output of a processing job, returning the full output once the job is complete.

This endpoint is especially useful for jobs where the complete output needs to be accessed at once. It provides a consolidated view of the results post-completion, ensuring that users can retrieve the entire output without the need to poll multiple times or manage partial results.

**Parameters:**
- **job_id** (string): The unique identifier of the job for which output is being requested. This ID is used to track the job's progress and aggregate its output.

**Returns:**
- **output** (Any): The aggregated output from the job, returned as a single entity once the job has concluded. The format of the output will depend on the nature of the job and how its results are structured.
"""

STATUS_DESCRIPTION = """
Checks the completion status of a processing job and returns its output if the job is complete.

This endpoint is invaluable for monitoring the progress of a job and obtaining the output only after the job has fully completed. It simplifies the process of querying job completion and retrieving results, eliminating the need for continuous polling or result aggregation.

**Parameters:**
- **job_id** (string): The unique identifier for the job being queried. This ID is used to track and assess the status of the job.

**Returns:**
- **status** (string): The completion status of the job, typically 'complete' or 'in progress'. This status indicates whether the job has finished processing and if the output is ready for retrieval.
- **output** (Any, optional): The final output of the job, provided if the job is complete. The format and structure of the output depend on the job's nature and the data processing involved.

**Note:** The availability of the `output` field is contingent on the job's completion status. If the job is still in progress, this field may be omitted or contain partial results, depending on the implementation.
"""


# ------------------------------ Initializations ----------------------------- #
job_list = JobsProgress()
heartbeat = Heartbeat()


# ------------------------------- Input Objects ------------------------------ #
@dataclass
class Job:
    """Represents a job."""

    id: str
    input: Union[dict, list, str, int, float, bool]


@dataclass
class TestJob:
    """Represents a test job.
    input can be any type of data.
    """

    id: Optional[str] = None
    input: Optional[Union[dict, list, str, int, float, bool]] = None
    webhook: Optional[str] = None


@dataclass
class DefaultRequest:
    """Represents a test input."""

    input: Dict[str, Any]
    webhook: Optional[str] = None


# ------------------------------ Output Objects ------------------------------ #
@dataclass
class JobOutput:
    """Represents the output of a job."""

    id: str
    status: str
    output: Optional[Union[dict, list, str, int, float, bool]] = None
    error: Optional[str] = None


@dataclass
class StreamOutput:
    """Stream representation of a job."""

    id: str
    status: str = "IN_PROGRESS"
    stream: Optional[Union[dict, list, str, int, float, bool]] = None
    error: Optional[str] = None


# ------------------------------ Webhook Sender ------------------------------ #
def _send_webhook(url: str, payload: Dict[str, Any]) -> bool:
    """
    Sends a webhook to the provided URL.

    Args:
        url (str): The URL to send the webhook to.
        payload (Dict[str, Any]): The JSON payload to send.

    Returns:
        bool: True if the request was successful, False otherwise.
    """
    with SyncClientSession() as session:
        try:
            response = session.post(url, json=payload, timeout=10)
            response.raise_for_status()  # Raises exception for 4xx/5xx responses
            return True
        except requests.RequestException as err:
            print(f"WEBHOOK | Request to {url} failed: {err}")
            return False


# ---------------------------------------------------------------------------- #
#                                  API Worker                                  #
# ---------------------------------------------------------------------------- #
class WorkerAPI:
    """Used to launch the FastAPI web server when the worker is running in API mode."""

    def __init__(self, config: Dict[str, Any]):
        """
        Initializes the WorkerAPI class.
        1. Starts the heartbeat thread.
        2. Initializes the FastAPI web server.
        3. Sets the handler for processing jobs.
        """
        # Start the heartbeat thread.
        heartbeat.start_ping()

        self.config = config

        tags_metadata = [
            {
                "name": "Synchronously Submit Request & Get Job Results",
                "description": "Endpoints for submitting job requests and getting the results.",
            },
            {
                "name": "Submit Job Requests",
                "description": "Endpoints for submitting job requests.",
            },
            {
                "name": "Check Job Results",
                "description": "Endpoints for checking the status of a job and getting the results.",
            },
        ]

        # Initialize the FastAPI web server.
        self.rp_app = FastAPI(
            title=TITLE,
            description=DESCRIPTION,
            version=runpod_version,
            docs_url="/",
            openapi_tags=tags_metadata,
        )

        # Create an APIRouter and add the route for processing jobs.
        api_router = APIRouter()

        # Docs Redirect /docs -> /
        api_router.add_api_route(
            "/docs", lambda: RedirectResponse(url="/"), include_in_schema=False
        )

        if RUNPOD_ENDPOINT_ID:
            api_router.add_api_route(
                f"/{RUNPOD_ENDPOINT_ID}/realtime", self._realtime, methods=["POST"]
            )

        # Simulation endpoints.
        api_router.add_api_route(
            "/run",
            self._sim_run,
            methods=["POST"],
            response_model_exclude_none=True,
            summary="Mimics the behavior of the run endpoint.",
            description=RUN_DESCRIPTION,
            tags=["Submit Job Requests"],
        )
        api_router.add_api_route(
            "/runsync",
            self._sim_runsync,
            methods=["POST"],
            response_model_exclude_none=True,
            summary="Mimics the behavior of the runsync endpoint.",
            description=RUNSYNC_DESCRIPTION,
            tags=["Synchronously Submit Request & Get Job Results"],
        )
        api_router.add_api_route(
            "/stream/{job_id}",
            self._sim_stream,
            methods=["POST"],
            response_model_exclude_none=True,
            summary="Mimics the behavior of the stream endpoint.",
            description=STREAM_DESCRIPTION,
            tags=["Check Job Results"],
        )
        api_router.add_api_route(
            "/status/{job_id}",
            self._sim_status,
            methods=["POST"],
            response_model_exclude_none=True,
            summary="Mimics the behavior of the status endpoint.",
            description=STATUS_DESCRIPTION,
            tags=["Check Job Results"],
        )

        # Include the APIRouter in the FastAPI application.
        self.rp_app.include_router(api_router)

    def start_uvicorn(self, api_host="localhost", api_port=8000, api_concurrency=1):
        """
        Starts the Uvicorn server.
        """
        uvicorn.run(
            self.rp_app,
            host=api_host,
            port=int(api_port),
            workers=int(api_concurrency),
            log_level=os.environ.get("UVICORN_LOG_LEVEL", "info"),
            access_log=False,
        )

    # ----------------------------- Realtime Endpoint ---------------------------- #
    async def _realtime(self, job: Job):
        """
        Performs model inference on the input data using the provided handler.
        If handler is not provided, returns an error message.
        """
        job_list.add(job.id)

        # Process the job using the provided handler, passing in the job input.
        job_results = await run_job(self.config["handler"], job.__dict__)

        job_list.remove(job.id)

        # Return the results of the job processing.
        return jsonable_encoder(job_results)

    # ---------------------------------------------------------------------------- #
    #                             Simulation Endpoints                             #
    # ---------------------------------------------------------------------------- #

    # ------------------------------------ run ----------------------------------- #
    async def _sim_run(self, job_request: DefaultRequest) -> JobOutput:
        """Development endpoint to simulate run behavior."""
        assigned_job_id = f"test-{uuid.uuid4()}"
        job_list.add({
            "id": assigned_job_id,
            "input": job_request.input,
            "webhook": job_request.webhook
        })
        return jsonable_encoder({"id": assigned_job_id, "status": "IN_PROGRESS"})

    # ---------------------------------- runsync --------------------------------- #
    async def _sim_runsync(self, job_request: DefaultRequest) -> JobOutput:
        """Development endpoint to simulate runsync behavior."""
        assigned_job_id = f"test-{uuid.uuid4()}"
        job = TestJob(id=assigned_job_id, input=job_request.input)

        if is_generator(self.config["handler"]):
            generator_output = run_job_generator(self.config["handler"], job.__dict__)
            job_output = {"output": []}
            async for stream_output in generator_output:
                job_output["output"].append(stream_output["output"])
        else:
            job_output = await run_job(self.config["handler"], job.__dict__)

        if job_output.get("error", None):
            return jsonable_encoder(
                {"id": job.id, "status": "FAILED", "error": job_output["error"]}
            )

        if job_request.webhook:
            thread = threading.Thread(
                target=_send_webhook,
                args=(job_request.webhook, job_output),
                daemon=True,
            )
            thread.start()

        return jsonable_encoder(
            {"id": job.id, "status": "COMPLETED", "output": job_output["output"]}
        )

    # ---------------------------------- stream ---------------------------------- #
    async def _sim_stream(self, job_id: str) -> StreamOutput:
        """Development endpoint to simulate stream behavior."""
        stashed_job = job_list.get(job_id)
        if stashed_job is None:
            return jsonable_encoder(
                {"id": job_id, "status": "FAILED", "error": "Job ID not found"}
            )

        job = TestJob(id=job_id, input=stashed_job.input)

        if is_generator(self.config["handler"]):
            generator_output = run_job_generator(self.config["handler"], job.__dict__)
            stream_accumulator = []
            async for stream_output in generator_output:
                stream_accumulator.append({"output": stream_output["output"]})
        else:
            return jsonable_encoder(
                {
                    "id": job_id,
                    "status": "FAILED",
                    "error": "Stream not supported, handler must be a generator.",
                }
            )

        job_list.remove(job.id)

        if stashed_job.webhook:
            thread = threading.Thread(
                target=_send_webhook,
                args=(stashed_job.webhook, stream_accumulator),
                daemon=True,
            )
            thread.start()

        return jsonable_encoder(
            {"id": job_id, "status": "COMPLETED", "stream": stream_accumulator}
        )

    # ---------------------------------- status ---------------------------------- #
    async def _sim_status(self, job_id: str) -> JobOutput:
        """Development endpoint to simulate status behavior."""
        stashed_job = job_list.get(job_id)
        if stashed_job is None:
            return jsonable_encoder(
                {"id": job_id, "status": "FAILED", "error": "Job ID not found"}
            )

        job = TestJob(id=stashed_job.id, input=stashed_job.input)

        if is_generator(self.config["handler"]):
            generator_output = run_job_generator(self.config["handler"], job.__dict__)
            job_output = {"output": []}
            async for stream_output in generator_output:
                job_output["output"].append(stream_output["output"])
        else:
            job_output = await run_job(self.config["handler"], job.__dict__)

        job_list.remove(job.id)

        if job_output.get("error", None):
            return jsonable_encoder(
                {"id": job_id, "status": "FAILED", "error": job_output["error"]}
            )

        if stashed_job.webhook:
            thread = threading.Thread(
                target=_send_webhook,
                args=(stashed_job.webhook, job_output),
                daemon=True,
            )
            thread.start()

        return jsonable_encoder(
            {"id": job_id, "status": "COMPLETED", "output": job_output["output"]}
        )



"""
runpod | cli | config.py

A collection of functions to set and validate configurations.
Configurations are TOML files located under ~/.runpod/
"""

import os
from pathlib import Path

import tomli as toml

CREDENTIAL_FILE = os.path.expanduser("~/.runpod/config.toml")


def set_credentials(api_key: str, profile: str = "default", overwrite=False) -> None:
    """
    Sets the user's credentials in ~/.runpod/config.toml
    If profile already exists user must use `update_credentials` instead.

    Args:
        api_key (str): The user's API key.
        profile (str): The profile to set the credentials for.

    --- File Structure ---

    [default]
    api_key = "RUNPOD_API_KEY"
    """
    os.makedirs(os.path.dirname(CREDENTIAL_FILE), exist_ok=True)
    Path(CREDENTIAL_FILE).touch(exist_ok=True)

    if not overwrite:
        with open(CREDENTIAL_FILE, "rb") as cred_file:
            if profile in toml.load(cred_file):
                raise ValueError(
                    "Profile already exists. Use `update_credentials` instead."
                )

    with open(CREDENTIAL_FILE, "w", encoding="UTF-8") as cred_file:
        cred_file.write("[" + profile + "]\n")
        cred_file.write('api_key = "' + api_key + '"\n')


def check_credentials(profile: str = "default"):
    """
    Checks if the credentials file exists and is valid.
    """
    if not os.path.exists(CREDENTIAL_FILE):
        return False, "~/.runpod/config.toml does not exist."

    # Check for default api_key
    try:
        with open(CREDENTIAL_FILE, "rb") as cred_file:
            config = toml.load(cred_file)

        if profile not in config:
            return False, f"~/.runpod/config.toml is missing {profile} profile."

        if "api_key" not in config[profile]:
            return (
                False,
                f"~/.runpod/config.toml is missing api_key for {profile} profile.",
            )

    except (TypeError, ValueError):
        return False, "~/.runpod/config.toml is not a valid TOML file."

    return True, None


def get_credentials(profile="default"):
    """
    Returns the credentials for the specified profile from ~/.runpod/config.toml
    """
    if not os.path.exists(CREDENTIAL_FILE):
        return None

    with open(CREDENTIAL_FILE, "rb") as cred_file:
        credentials = toml.load(cred_file)

    if profile not in credentials:
        return None

    return credentials[profile]



"""
Provides some of the torch.cuda functionality without requiring torch.
"""

import subprocess


def is_available():
    """
    Returns True if CUDA is available, False otherwise.
    """
    try:
        output = subprocess.check_output("nvidia-smi", shell=True)
        if "NVIDIA-SMI" in output.decode():
            return True
    except Exception:  # pylint: disable=broad-except
        pass
    return False



"""
RunPod | Tests | CLI | Commands
"""

import unittest
from unittest.mock import patch

from click.testing import CliRunner

from runpod.cli.entry import runpod_cli


class TestCommands(unittest.TestCase):
    """A collection of tests for the CLI commands."""

    def setUp(self):
        self.runner = CliRunner()

    def test_config_wizard(self):
        """Tests the config command."""
        with patch("click.echo") as mock_echo, patch(
            "runpod.cli.groups.config.commands.set_credentials"
        ) as mock_set_credentials, patch(
            "runpod.cli.groups.config.commands.check_credentials"
        ) as mock_check_creds, patch(
            "click.confirm", return_value=True
        ) as mock_confirm, patch(
            "click.prompt", return_value="KEY"
        ) as mock_prompt:

            # Assuming credentials aren't set (doesn't prompt for overwrite)
            mock_check_creds.return_value = (False, None)

            # Successful Call with Direct Key
            result = self.runner.invoke(
                runpod_cli, ["config", "--profile", "test", "KEY"]
            )
            assert result.exit_code == 0
            mock_set_credentials.assert_called_with("KEY", "test", overwrite=True)
            assert mock_echo.call_count == 1

            # Successful Call with Prompted Key (since direct key isn't provided)
            result = self.runner.invoke(runpod_cli, ["config", "--profile", "test"])
            assert result.exit_code == 0
            mock_set_credentials.assert_called_with("KEY", "test", overwrite=True)
            mock_prompt.assert_called_with(
                "    > RunPod API Key", hide_input=False, confirmation_prompt=False
            )  # pylint: disable=line-too-long

            # Simulating existing credentials, prompting for overwrite
            mock_check_creds.return_value = (True, None)
            result = self.runner.invoke(runpod_cli, ["config", "--profile", "test"])
            mock_confirm.assert_called_with(
                "Credentials already set for profile: test. Overwrite?", abort=True
            )

            # Unsuccessful Call
            mock_set_credentials.side_effect = ValueError()
            result = self.runner.invoke(
                runpod_cli, ["config", "--profile", "test", "KEY"]
            )
            assert result.exit_code == 1

    def test_check_flag(self):
        """Tests the --check flag."""
        with patch(
            "runpod.cli.groups.config.commands.check_credentials"
        ) as mock_check_creds:
            # Assuming credentials are set
            mock_check_creds.return_value = (True, None)
            result = self.runner.invoke(
                runpod_cli, ["config", "--check", "--profile", "test"]
            )
            assert result.exit_code == 0

            # Assuming credentials aren't set
            mock_check_creds.return_value = (False, "Credentials not found.")
            result = self.runner.invoke(
                runpod_cli, ["config", "--check", "--profile", "test"]
            )
            assert result.exit_code == 1

    def test_output_messages(self):
        """Tests the output messages for the config command."""
        with patch("click.echo") as mock_echo, patch(
            "runpod.cli.groups.config.commands.set_credentials"
        ) as mock_set_credentials, patch(
            "runpod.cli.groups.config.commands.check_credentials",
            return_value=(False, None),
        ) as mock_check_creds:  # pylint: disable=line-too-long
            result = self.runner.invoke(
                runpod_cli, ["config", "KEY", "--profile", "test"]
            )
            mock_set_credentials.assert_called_with("KEY", "test", overwrite=True)
            mock_echo.assert_any_call(
                "Credentials set for profile: test in ~/.runpod/config.toml"
            )
            assert result.exit_code == 0
            assert mock_check_creds.call_count == 1

    def test_api_key_prompt(self):
        """Tests the API key prompt."""
        with patch("click.prompt", return_value="KEY") as mock_prompt:
            result = self.runner.invoke(runpod_cli, ["config", "--profile", "test"])
            mock_prompt.assert_called_with(
                "    > RunPod API Key", hide_input=False, confirmation_prompt=False
            )  # pylint: disable=line-too-long
            assert result.exit_code == 0



"""'
GraphQL wrapper for the RunPod API
"""

import time

import runpod

runpod.api_key = "YOUR_RUNPOD_API_KEY"

# Get all GPUs
gpus = runpod.get_gpus()
print(gpus)

# Get a specific GPU
gpu = runpod.get_gpu("NVIDIA GeForce RTX 3070")
print(gpu)

# Create a pod
pod = runpod.create_pod("test", "runpod/stack", "NVIDIA GeForce RTX 3070")
print(pod)

# Pause while the pod is being created
print("Waiting for pod to be created...")
time.sleep(10)

# Stop a pod
pod = runpod.stop_pod(pod["id"])
print(pod)

# Pause while the pod is being stopped
print("Waiting for pod to be stopped...")
time.sleep(10)

# Resume a pod
pod = runpod.resume_pod(pod["id"], 1)
print(pod)

# Pause while the pod is being resumed
print("Waiting for pod to be resumed...")
time.sleep(10)

# Terminate a pod
runpod.terminate_pod(pod["id"])



"""
Job related helpers.
"""

import inspect
import json
import os
import traceback
from typing import Any, AsyncGenerator, Callable, Dict, Optional, Union, List

import aiohttp

from runpod.http_client import ClientSession, TooManyRequests
from runpod.serverless.modules.rp_logger import RunPodLogger

from ...version import __version__ as runpod_version
from ..utils import rp_debugger
from .rp_handler import is_generator
from .rp_http import send_result, stream_result
from .rp_tips import check_return_size
from .worker_state import WORKER_ID, REF_COUNT_ZERO, JobsProgress

JOB_GET_URL = str(os.environ.get("RUNPOD_WEBHOOK_GET_JOB")).replace("$ID", WORKER_ID)

log = RunPodLogger()
job_progress = JobsProgress()


def _job_get_url(batch_size: int = 1):
    """
    Prepare the URL for making a 'get' request to the serverless API (sls).

    This function constructs the appropriate URL for sending a 'get' request to the serverless API,
    ensuring that the request will be correctly routed and processed by the API.

    Returns:
        str: The prepared URL for the 'get' request to the serverless API.
    """

    if batch_size > 1:
        job_take_url = JOB_GET_URL.replace("/job-take/", "/job-take-batch/")
        job_take_url += f"&batch_size={batch_size}"
    else:
        job_take_url = JOB_GET_URL

    job_in_progress = "1" if job_progress.get_job_list() else "0"
    job_take_url += f"&job_in_progress={job_in_progress}"

    log.debug(f"rp_job | get_job: {job_take_url}")
    return job_take_url


async def get_job(
    session: ClientSession, num_jobs: int = 1
) -> Optional[List[Dict[str, Any]]]:
    """
    Get a job from the job-take API.

    `num_jobs = 1` will query the legacy singular job-take API.

    `num_jobs > 1` will query the batch job-take API.

    Args:
        session (ClientSession): The aiohttp ClientSession to use for the request.
        num_jobs (int): The number of jobs to get.
    """
    async with session.get(_job_get_url(num_jobs)) as response:
        log.debug(f"rp_job | Response: {type(response).__name__} {response.status}")

        if response.status == 204:
            log.debug("rp_job | Received 204 status, no jobs.")
            return

        if response.status == 400:
            log.debug("rp_job | Received 400 status, expected when FlashBoot is enabled.")
            return

        if response.status == 429:
            raise TooManyRequests(
                response.request_info,
                response.history,
                status=response.status,
                message=response.reason
            )

        # All other errors should raise an exception
        response.raise_for_status()

        # Verify if the content type is JSON
        if response.content_type != "application/json":
            log.debug(f"rp_job | Unexpected content type: {response.content_type}")
            return

        # Check if there is a non-empty content to parse
        if response.content_length == 0:
            log.debug("rp_job | No content to parse.")
            return

        try:
            jobs = await response.json()
            log.debug("rp_job | Received Job(s)")
        except aiohttp.ContentTypeError:
            log.debug(f"rp_job | Response content is not valid JSON. {response.content}")
            return
        except ValueError as json_error:
            log.debug(f"rp_job | Failed to parse JSON response: {json_error}")
            return

        # legacy job-take API
        if isinstance(jobs, dict):
            if "id" not in jobs or "input" not in jobs:
                raise Exception("Job has missing field(s): id or input.")
            return [jobs]

        # batch job-take API
        if isinstance(jobs, list):
            return jobs


async def handle_job(session: ClientSession, config: Dict[str, Any], job) -> dict:
    if is_generator(config["handler"]):
        is_stream = True
        generator_output = run_job_generator(config["handler"], job)
        log.debug("Handler is a generator, streaming results.", job["id"])

        job_result = {"output": []}
        async for stream_output in generator_output:
            log.debug(f"Stream output: {stream_output}", job["id"])

            if type(stream_output.get("output")) == dict:
                if stream_output["output"].get("error"):
                    stream_output = {"error": str(stream_output["output"]["error"])}

            if stream_output.get("error"):
                job_result = stream_output
                break

            if config.get("return_aggregate_stream", False):
                job_result["output"].append(stream_output["output"])

            await stream_result(session, stream_output, job)
    else:
        is_stream = False
        job_result = await run_job(config["handler"], job)

    # If refresh_worker is set, pod will be reset after job is complete.
    if config.get("refresh_worker", False):
        log.info("refresh_worker flag set, stopping pod after job.", job["id"])
        job_result["stopPod"] = True

    # If rp_debugger is set, debugger output will be returned.
    if config["rp_args"].get("rp_debugger", False) and isinstance(job_result, dict):
        job_result["output"]["rp_debugger"] = rp_debugger.get_debugger_output()
        log.debug("rp_debugger | Flag set, returning debugger output.", job["id"])

        # Calculate ready delay for the debugger output.
        ready_delay = (config["reference_counter_start"] - REF_COUNT_ZERO) * 1000
        job_result["output"]["rp_debugger"]["ready_delay_ms"] = ready_delay
    else:
        log.debug("rp_debugger | Flag not set, skipping debugger output.", job["id"])
        rp_debugger.clear_debugger_output()

    # Send the job result back to JOB_DONE_URL
    await send_result(session, job_result, job, is_stream=is_stream)


async def run_job(handler: Callable, job: Dict[str, Any]) -> Dict[str, Any]:
    """
    Run the job using the handler.

    Args:
        handler (Callable): The handler function to use.
        job (Dict[str, Any]): The job to run.

    Returns:
        Dict[str, Any]: The result of running the job.
    """
    log.info("Started.", job["id"])
    run_result = {}

    try:
        handler_return = handler(job)
        job_output = (
            await handler_return
            if inspect.isawaitable(handler_return)
            else handler_return
        )

        log.debug(f"Handler output: {job_output}", job["id"])

        if isinstance(job_output, dict):
            error_msg = job_output.pop("error", None)
            refresh_worker = job_output.pop("refresh_worker", None)
            run_result["output"] = job_output

            if error_msg:
                run_result["error"] = error_msg
            if refresh_worker:
                run_result["stopPod"] = True

        elif isinstance(job_output, bool):
            run_result = {"output": job_output}

        else:
            run_result = {"output": job_output}

        if run_result.get("output") == {}:
            run_result.pop("output")

        check_return_size(run_result)  # Checks the size of the return body.

    except Exception as err:
        error_info = {
            "error_type": str(type(err)),
            "error_message": str(err),
            "error_traceback": traceback.format_exc(),
            "hostname": os.environ.get("RUNPOD_POD_HOSTNAME", "unknown"),
            "worker_id": os.environ.get("RUNPOD_POD_ID", "unknown"),
            "runpod_version": runpod_version,
        }

        log.error("Captured Handler Exception", job["id"])
        log.error(json.dumps(error_info, indent=4))
        run_result = {"error": json.dumps(error_info)}

    finally:
        log.debug(f"run_job return: {run_result}", job["id"])

    return run_result


async def run_job_generator(
    handler: Callable, job: Dict[str, Any]
) -> AsyncGenerator[Dict[str, Union[str, Any]], None]:
    """
    Run generator job used to stream output.
    Yields output partials from the generator.
    """
    is_async_gen = inspect.isasyncgenfunction(handler)
    log.debug(
        "Using Async Generator" if is_async_gen else "Using Standard Generator",
        job["id"],
    )

    try:
        job_output = handler(job)

        if is_async_gen:
            async for output_partial in job_output:
                log.debug(f"Async Generator output: {output_partial}", job["id"])
                yield {"output": output_partial}
        else:
            for output_partial in job_output:
                log.debug(f"Generator output: {output_partial}", job["id"])
                yield {"output": output_partial}

    except Exception as err:
        log.error(err, job["id"])
        yield {"error": f"handler: {str(err)} \ntraceback: {traceback.format_exc()}"}
    finally:
        log.info("Finished running generator.", job["id"])



"""Retrieve handler info. """

import inspect
from typing import Callable


def is_generator(handler: Callable) -> bool:
    """Check if handler is a generator function."""
    return inspect.isgeneratorfunction(handler) or inspect.isasyncgenfunction(handler)



""" Allows runpod to be imported as a module. """

import logging
import os

from . import serverless
from .api.ctl_commands import (
    create_container_registry_auth,
    create_endpoint,
    create_pod,
    create_template,
    delete_container_registry_auth,
    get_endpoints,
    get_gpu,
    get_gpus,
    get_pod,
    get_pods,
    get_user,
    resume_pod,
    stop_pod,
    terminate_pod,
    update_container_registry_auth,
    update_endpoint_template,
    update_user_settings,
)
from .cli.groups.config.functions import (
    check_credentials,
    get_credentials,
    set_credentials,
)
from .endpoint import AsyncioEndpoint, AsyncioJob, Endpoint
from .serverless.modules.rp_logger import RunPodLogger
from .version import __version__

# ------------------------------- Config Paths ------------------------------- #
SSH_KEY_PATH = os.path.expanduser("~/.runpod/ssh")


profile = "default"  # pylint: disable=invalid-name

_credentials = get_credentials(profile)
if _credentials is not None:
    api_key = _credentials["api_key"]  # pylint: disable=invalid-name
else:
    api_key = None  # pylint: disable=invalid-name

endpoint_url_base = os.environ.get(
    "RUNPOD_ENDPOINT_BASE_URL", "https://api.runpod.ai/v2"
)  # pylint: disable=invalid-name


# --------------------------- Force Logging Levels --------------------------- #
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("paramiko").setLevel(logging.WARNING)
logging.getLogger("uvicorn").setLevel(logging.WARNING)



"""Tests for the endpoint mutation generation."""

import unittest

from runpod.api.mutations.endpoints import generate_endpoint_mutation


class TestGenerateEndpointMutation(unittest.TestCase):
    """Tests for the endpoint mutation generation."""

    def test_required_fields(self):
        """Test the required fields."""
        result = generate_endpoint_mutation("test_name", "test_template_id")
        self.assertIn('name: "test_name"', result)
        self.assertIn('templateId: "test_template_id"', result)
        self.assertIn('gpuIds: "AMPERE_16"', result)  # Default value
        self.assertIn('networkVolumeId: ""', result)  # Default value
        self.assertIn('locations: ""', result)  # Default value

    def test_all_fields(self):
        """Test all the fields."""
        result = generate_endpoint_mutation(
            "test_name",
            "test_template_id",
            "AMPERE_20",
            "test_volume_id",
            "US_WEST",
            10,
            "WORKER_COUNT",
            5,
            2,
            4,
            True,
        )
        self.assertIn('name: "test_name-fb"', result)
        self.assertIn('templateId: "test_template_id"', result)
        self.assertIn('gpuIds: "AMPERE_20"', result)
        self.assertIn('networkVolumeId: "test_volume_id"', result)
        self.assertIn('locations: "US_WEST"', result)
        self.assertIn("idleTimeout: 10", result)
        self.assertIn('scalerType: "WORKER_COUNT"', result)
        self.assertIn("scalerValue: 5", result)
        self.assertIn("workersMin: 2", result)
        self.assertIn("workersMax: 4", result)



"""
Unit tests for the debugger utility functions.
"""

import importlib
import time
import unittest
from unittest.mock import patch

from runpod.serverless.utils import rp_debugger
from runpod.serverless.utils.rp_debugger import (
    Checkpoints,
    FunctionTimer,
    LineTimer,
    clear_debugger_output,
    get_debugger_output,
)


class TestDebugger(unittest.TestCase):
    """Unit tests for the debugger utility functions."""

    def setUp(self):
        self.checkpoints = Checkpoints()
        self.checkpoints.clear()

    @patch("runpod.serverless.utils.rp_debugger.cpuinfo.get_cpu_info")
    def test_key_error(self, mock_get_cpu_info):
        """
        Test that a KeyError is raised when an invalid key is used.
        """
        mock_get_cpu_info.side_effect = KeyError("Test Error")

        importlib.reload(rp_debugger)

        assert mock_get_cpu_info.called
        self.assertEqual(rp_debugger.PROCESSOR, "Unable to get processor info.")

    def test_checkpoints(self):
        """
        Test that checkpoints are added and stopped correctly.
        """
        self.checkpoints.add("checkpoint1")
        self.checkpoints.start("checkpoint1")
        time.sleep(0.1)
        self.checkpoints.stop("checkpoint1")

        output = get_debugger_output()["timestamps"]

        self.assertEqual(len(output), 1)
        self.assertEqual(output[0]["name"], "checkpoint1")
        self.assertGreater(output[0]["duration_ms"], 0)

    def test_get_checkpoints(self):
        """
        Test that checkpoints are added and stopped correctly.
        """
        self.checkpoints.add("checkpoint1")
        checkpoint_list = self.checkpoints.get_checkpoints()

        self.assertEqual(len(checkpoint_list), 0)

    def test_checkpoints_exception(self):
        """
        Test that a KeyError is raised when an invalid checkpoint is used.
        """
        self.assertRaises(KeyError, self.checkpoints.start, "nonexistent")
        self.assertRaises(KeyError, self.checkpoints.stop, "nonexistent")

        self.checkpoints.add("checkpoint2")
        self.assertRaises(KeyError, self.checkpoints.add, "checkpoint2")
        self.assertRaises(KeyError, self.checkpoints.stop, "checkpoint2")

    def test_line_timer(self):
        """
        Test that the line timer works correctly.
        """
        with LineTimer("line_timer"):
            time.sleep(0.1)

        output = get_debugger_output()["timestamps"]

        self.assertEqual(len(output), 1)
        self.assertEqual(output[0]["name"], "line_timer")
        self.assertGreater(output[0]["duration_ms"], 0)

    def test_function_timer(self):
        """
        Test that the function timer works correctly.
        """

        @FunctionTimer
        def func_to_time():
            time.sleep(0.1)

        func_to_time()

        output = get_debugger_output()["timestamps"]

        self.assertEqual(len(output), 1)
        self.assertEqual(output[0]["name"], "func_to_time")
        self.assertGreater(output[0]["duration_ms"], 0)

    def test_clear_debugger_output(self):
        """
        Test that the debugger output is cleared correctly.
        """
        self.checkpoints.add("checkpoint1")
        self.checkpoints.start("checkpoint1")

        # Check that non-stopped checkpoints are not returned
        checkpoint_list = self.checkpoints.get_checkpoints()
        self.assertEqual(len(checkpoint_list), 0)

        self.checkpoints.stop("checkpoint1")

        checkpoint_list = self.checkpoints.get_checkpoints()
        self.assertEqual(len(checkpoint_list), 1)

        clear_debugger_output()

        checkpoint_list = self.checkpoints.get_checkpoints()
        self.assertEqual(len(checkpoint_list), 0)


if __name__ == "__main__":
    unittest.main()



"""
PodWorker | modules | download.py

Called when inputs are images or zip files.
Downloads them into a temporary directory called "input_objects".
This directory is cleaned up after the job is complete.
"""

import os
import re
import uuid
import zipfile
from concurrent.futures import ThreadPoolExecutor
from email import message_from_string
from typing import List, Union
from urllib.parse import urlparse

import backoff
from requests import RequestException

from runpod.http_client import SyncClientSession

HEADERS = {"User-Agent": "runpod-python/0.0.0 (https://runpod.io; support@runpod.io)"}


def calculate_chunk_size(file_size: int) -> int:
    """
    Calculates the chunk size based on the file size.
    """
    if file_size <= 1024 * 1024:  # 1 MB
        return 1024  # 1 KB
    if file_size <= 1024 * 1024 * 1024:  # 1 GB
        return 1024 * 1024  # 1 MB

    return 1024 * 1024 * 10  # 10 MB


def download_files_from_urls(job_id: str, urls: Union[str, List[str]]) -> List[str]:
    """
    Accepts a single URL or a list of URLs and downloads the files.
    Returns the list of downloaded file absolute paths.
    Saves the files in a directory called "downloaded_files" in the job directory.
    """
    download_directory = os.path.abspath(
        os.path.join("jobs", job_id, "downloaded_files")
    )
    os.makedirs(download_directory, exist_ok=True)

    @backoff.on_exception(backoff.expo, RequestException, max_tries=3)
    def download_file(url: str, path_to_save: str) -> str:
        with SyncClientSession().get(
            url, headers=HEADERS, stream=True, timeout=5
        ) as response:
            response.raise_for_status()
            content_disposition = response.headers.get("Content-Disposition")
            file_extension = ""
            if content_disposition:
                msg = message_from_string(f"Content-Disposition: {content_disposition}")
                params = dict(msg.items())
                file_extension = os.path.splitext(params.get("filename", ""))[1]

            # If no extension could be determined from 'Content-Disposition', get it from the URL
            if not file_extension:
                file_extension = os.path.splitext(urlparse(url).path)[1]

            file_size = int(response.headers.get("Content-Length", 0))
            chunk_size = calculate_chunk_size(file_size)

            # write the content in chunks to the file
            with open(path_to_save + file_extension, "wb") as file_path:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    if chunk:  # filter out keep-alive chunks
                        file_path.write(chunk)

            return file_extension

    def download_file_to_path(url: str) -> str:
        if url is None:
            return None

        file_name = f"{uuid.uuid4()}"
        output_file_path = os.path.join(download_directory, file_name)

        try:
            file_extension = download_file(url, output_file_path)
        except RequestException as err:
            print(f"Failed to download {url}: {err}")
            return None

        return os.path.abspath(f"{output_file_path}{file_extension}")

    if isinstance(urls, str):
        urls = [urls]

    with ThreadPoolExecutor() as executor:
        downloaded_files = list(executor.map(download_file_to_path, urls))

    return downloaded_files


def file(file_url: str) -> dict:
    """
    Downloads a single file from a given URL, file is given a random name.
    First checks if the content-disposition header is set, if so, uses the file name from there.
    If the file is a zip file, it is extracted into a directory with the same name.

    Returns an object that contains:
    - The absolute path to the downloaded file
    - File type
    - Original file name
    """
    os.makedirs("job_files", exist_ok=True)

    download_response = SyncClientSession().get(file_url, headers=HEADERS, timeout=30)

    original_file_name = []
    if "Content-Disposition" in download_response.headers.keys():
        original_file_name = re.findall(
            "filename=(.+)", download_response.headers["Content-Disposition"]
        )

    if len(original_file_name) > 0:
        original_file_name = original_file_name[0]
    else:
        download_path = urlparse(file_url).path
        original_file_name = os.path.basename(download_path)

    file_type = os.path.splitext(original_file_name)[1].replace(".", "")

    file_name = f"{uuid.uuid4()}"

    output_file_path = os.path.join("job_files", f"{file_name}.{file_type}")
    with open(output_file_path, "wb") as output_file:
        output_file.write(download_response.content)

    if file_type == "zip":
        unziped_directory = os.path.join("job_files", file_name)
        os.makedirs(unziped_directory, exist_ok=True)
        with zipfile.ZipFile(output_file_path, "r") as zip_ref:
            zip_ref.extractall(unziped_directory)
        unziped_directory = os.path.abspath(unziped_directory)
    else:
        unziped_directory = None

    return {
        "file_path": os.path.abspath(output_file_path),
        "type": file_type,
        "original_name": original_file_name,
        "extracted_path": unziped_directory,
    }



"""Tests for the project helpers."""

import unittest
from unittest.mock import mock_open, patch

import click

from runpod import error as rp_error
from runpod.cli.groups.project.helpers import (
    attempt_pod_launch,
    copy_template_files,
    get_project_endpoint,
    get_project_pod,
    load_project_config,
    validate_project_name,
)


class TestHelpers(unittest.TestCase):
    """Test the project helpers."""

    def test_validate_project_name_valid(self):
        """Test the validate_project_name function with valid input."""
        name = "validProjectName"
        result = validate_project_name(name)
        self.assertEqual(result, name)

    def test_validate_project_name_invalid(self):
        """Test the validate_project_name function with invalid input."""
        name = "invalid:name"
        with self.assertRaises(click.BadParameter):
            validate_project_name(name)

    @patch("runpod.cli.groups.project.helpers.get_pods")
    def test_get_project_pod_exists(self, mock_get_pods):
        """Test the get_project_pod function when the project pod exists."""
        mock_get_pods.return_value = [{"name": "test-1234", "id": "pod_id"}]
        result = get_project_pod("1234")
        self.assertEqual(result, "pod_id")

    @patch("runpod.cli.groups.project.helpers.get_pods")
    def test_get_project_pod_not_exists(self, mock_get_pods):
        """Test the get_project_pod function when the project pod doesn't exist."""
        mock_get_pods.return_value = [{"name": "another-5678", "id": "another_pod_id"}]
        result = get_project_pod("1234")
        self.assertIsNone(result)

    @patch("runpod.cli.groups.project.helpers.get_endpoints")
    def test_get_project_endpoint_exists(self, mock_get_endpoints):
        """Test the get_project_endpoint function when the project endpoint exists."""
        mock_get_endpoints.return_value = []
        assert get_project_endpoint("1234") is None

        mock_get_endpoints.return_value = [{"name": "test-1234", "id": "endpoint_id"}]
        result = get_project_endpoint("1234")
        self.assertEqual(result, {"name": "test-1234", "id": "endpoint_id"})

    @patch("os.listdir")
    @patch("os.path.isdir", return_value=False)
    @patch("shutil.copy2")
    def test_copy_template_files(self, mock_copy, mock_isdir, mock_listdir):
        """Test the copy_template_files function."""
        mock_listdir.return_value = ["file1.txt", "file2.txt"]
        copy_template_files("/template", "/destination")
        self.assertEqual(mock_copy.call_count, 2)
        assert mock_isdir.called

    @patch("os.listdir")
    @patch("os.path.isdir", return_value=True)
    @patch("shutil.copytree")
    def test_copy_template_files_dir(self, mock_copy, mock_isdir, mock_listdir):
        """Test the copy_template_files function."""
        mock_listdir.return_value = ["file1.txt", "file2.txt"]
        copy_template_files("/template", "/destination")
        self.assertEqual(mock_copy.call_count, 2)
        assert mock_isdir.called

    @patch("runpod.cli.groups.project.helpers.create_pod")
    def test_attempt_pod_launch_success(self, mock_create_pod):
        """Test the attempt_pod_launch function when it succeeds."""
        mock_create_pod.return_value = "pod_id"
        config = {
            "project": {
                "name": "test",
                "uuid": "1234",
                "base_image": "base_image",
                "gpu_types": ["gpu_type"],
                "gpu_count": "1",
                "ports": "ports",
                "storage_id": "storage_id",
                "volume_mount_path": "volume_mount_path",
                "container_disk_size_gb": "1",
            }
        }
        environment_variables = {"key": "value"}
        result = attempt_pod_launch(config, environment_variables)
        self.assertEqual(result, "pod_id")

        mock_create_pod.side_effect = rp_error.QueryError("error")
        assert attempt_pod_launch(config, environment_variables) is None

    @patch("os.path.exists", return_value=True)
    @patch("builtins.open", new_callable=mock_open, read_data="[project]\nname='test'")
    def test_load_project_config(self, mock_file, mock_exists):
        """Test the load_project_config function."""
        config = load_project_config()
        self.assertEqual(config["project"]["name"], "test")
        assert mock_exists.called
        assert mock_file.called

        with patch("os.path.exists", return_value=False), self.assertRaises(
            FileNotFoundError
        ):
            load_project_config()



""" Example of creating a container registry auth with the Runpod API. """

import runpod

# Set your global API key with `runpod config` or uncomment the line below:
# runpod.api_key = "YOUR_RUNPOD_API_KEY"

try:
    new_container_registry_auth = runpod.create_container_registry_auth(
        name="test-container-registry-auth-name",
        username="test-username",
        password="test-password",
    )

    print(new_container_registry_auth)

except runpod.error.QueryError as err:
    print(err)
    print(err.query)
